/*
 * Name:    d_network_mce.c
 *
 * Purpose: MX MCE driver to support MX network multichannel encoders.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2000-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_mce.h"
#include "mx_net.h"
#include "d_network_mce.h"

/* Initialize the MCE driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_mce_record_function_list = {
	mxd_network_mce_initialize_type,
	mxd_network_mce_create_record_structures,
	mxd_network_mce_finish_record_initialization,
	mxd_network_mce_delete_record,
	NULL,
	mxd_network_mce_read_parms_from_hardware,
	mxd_network_mce_write_parms_to_hardware,
	mxd_network_mce_open,
	mxd_network_mce_close
};

MX_MCE_FUNCTION_LIST mxd_network_mce_mce_function_list = {
	mxd_network_mce_get_overflow_status,
	mxd_network_mce_reset_overflow_status,
	mxd_network_mce_read,
	mxd_network_mce_get_current_num_values
};

/* MCS encoder data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_network_mce_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCE_STANDARD_FIELDS,
	MXD_NETWORK_MCE_STANDARD_FIELDS
};

long mxd_network_mce_num_record_fields
		= sizeof( mxd_network_mce_record_field_defaults )
		  / sizeof( mxd_network_mce_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_mce_rfield_def_ptr
			= &mxd_network_mce_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_network_mce_get_pointers( MX_MCE *mce,
			MX_NETWORK_MCE **network_mce,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_mce_get_pointers()";

	if ( mce == (MX_MCE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCE pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( network_mce == (MX_NETWORK_MCE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_NETWORK_MCE pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*network_mce = (MX_NETWORK_MCE *) (mce->record->record_type_struct);

	if ( *network_mce == (MX_NETWORK_MCE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_NETWORK_MCE pointer for record '%s' is NULL.",
			mce->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_network_mce_initialize_type( long record_type )
{
	long num_record_fields;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long maximum_num_values_varargs_cookie;
	mx_status_type mx_status;

	mx_status = mx_mce_initialize_type( record_type,
					&num_record_fields,
					&record_field_defaults,
					&maximum_num_values_varargs_cookie );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_mce_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mce_create_record_structures()";

	MX_MCE *mce;
	MX_NETWORK_MCE *network_mce;

	/* Allocate memory for the necessary structures. */

	mce = (MX_MCE *) malloc( sizeof(MX_MCE) );

	if ( mce == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCE structure." );
	}

	network_mce = (MX_NETWORK_MCE *) malloc( sizeof(MX_NETWORK_MCE) );

	if ( network_mce == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_MCE structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mce;
	record->record_type_struct = network_mce;
	record->class_specific_function_list
			= &mxd_network_mce_mce_function_list;

	mce->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_network_mce_finish_record_initialization()";

	MX_MCE *mce;

	mce = (MX_MCE *) record->record_class_struct;

	if ( mce == (MX_MCE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCE pointer for record '%s' is NULL.",
			record->name );
	}

	mce->encoder_type = MXT_MCE_ABSOLUTE_ENCODER;

	mce->current_num_values = mce->maximum_num_values;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_get_overflow_status( MX_MCE *mce )
{
	mce->overflow_set = FALSE;
	mce->underflow_set = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_reset_overflow_status( MX_MCE *mce )
{
	mce->overflow_set = FALSE;
	mce->underflow_set = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_read( MX_MCE *mce )
{
	const char fname[] = "mxd_network_mce_read()";

	MX_NETWORK_MCE *network_mce;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[1];
	long num_values;
	mx_status_type mx_status;

	mx_status = mxd_network_mce_get_pointers( mce, &network_mce, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCE '%s'",
			fname, mce->record->name));

	/* First get the current number of encoder values. */

	sprintf( rfname, "%s.current_num_values",
				network_mce->remote_record_name );

	MX_DEBUG( 2,("%s: reading '%s'.", fname, rfname));

	mx_status = mx_get( network_mce->server_record,
				rfname, MXFT_LONG, &num_values );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: num_values = %ld", fname, num_values));

	MX_DEBUG( 2,("%s: mce->maximum_num_values = %ld",
				fname, mce->maximum_num_values));

	if ( num_values > mce->maximum_num_values ) {
		mce->current_num_values = mce->maximum_num_values;
	} else {
		mce->current_num_values = num_values;
	}

	MX_DEBUG( 2,("%s: mce->current_num_values = %ld",
			fname, mce->current_num_values ));

	/* Now transfer the MCE contents. */

	sprintf( rfname, "%s.value_array", network_mce->remote_record_name );

	MX_DEBUG( 2,("%s: reading '%s'.", fname, rfname));

	dimension_array[0] = mce->current_num_values;

	mx_status = mx_get_array( network_mce->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				mce->value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: data received for '%s'.", fname, rfname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_mce_get_current_num_values( MX_MCE *mce )
{
	const char fname[] = "mxd_network_mce_get_current_num_values()";

	MX_NETWORK_MCE *network_mce;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long num_values;
	mx_status_type mx_status;

	mx_status = mxd_network_mce_get_pointers( mce, &network_mce, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCE '%s'",
			fname, mce->record->name));

	sprintf( rfname, "%s.current_num_values",
				network_mce->remote_record_name );

	MX_DEBUG( 2,("%s: reading '%s'.", fname, rfname));

	mx_status = mx_get( network_mce->server_record,
				rfname, MXFT_LONG, &num_values );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: num_values = %ld", fname, num_values));

	MX_DEBUG( 2,("%s: mce->maximum_num_values = %ld",
				fname, mce->maximum_num_values));

	if ( num_values > mce->maximum_num_values ) {
		mce->current_num_values = mce->maximum_num_values;
	} else {
		mce->current_num_values = num_values;
	}

	MX_DEBUG( 2,("%s: mce->current_num_values = %ld",
			fname, mce->current_num_values ));

	return MX_SUCCESSFUL_RESULT;
}

