/*
 * Name:    d_pmac.c
 *
 * Purpose: MX driver for Delta Tau PMAC controlled motors.
 *
 *          Note that this driver performs its work using the PMAC jog
 *          commands, rather than using a motion program.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "d_pmac.h"

/* ============ Motor channels ============ */

MX_RECORD_FUNCTION_LIST mxd_pmac_record_function_list = {
	mxd_pmac_initialize_type,
	mxd_pmac_create_record_structures,
	mxd_pmac_finish_record_initialization,
	mxd_pmac_delete_record,
	mxd_pmac_print_structure,
	mxd_pmac_read_parms_from_hardware,
	mxd_pmac_write_parms_to_hardware,
	mxd_pmac_open,
	mxd_pmac_close,
	NULL,
	mxd_pmac_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_motor_function_list = {
	mxd_pmac_motor_is_busy,
	mxd_pmac_move_absolute,
	mxd_pmac_get_position,
	mxd_pmac_set_position,
	mxd_pmac_soft_abort,
	mxd_pmac_immediate_abort,
	mxd_pmac_positive_limit_hit,
	mxd_pmac_negative_limit_hit,
	mxd_pmac_find_home_position,
	mxd_pmac_constant_velocity_move,
	mxd_pmac_get_parameter,
	mxd_pmac_set_parameter,
	mxd_pmac_simultaneous_start
};

#define PMAC_DEBUG	FALSE

MX_RECORD_FIELD_DEFAULTS mxd_pmac_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_STANDARD_FIELDS
};

long mxd_pmac_num_record_fields
		= sizeof( mxd_pmac_record_field_defaults )
			/ sizeof( mxd_pmac_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_rfield_def_ptr
			= &mxd_pmac_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_pmac_get_pointers( MX_MOTOR *motor,
			MX_PMAC_MOTOR **pmac_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_pmac_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac_motor == (MX_PMAC_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*pmac_motor = (MX_PMAC_MOTOR *) (motor->record->record_type_struct);

	if ( *pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_pmac_motor_get_pmac_pointer( MX_PMAC_MOTOR *pmac_motor,
			MX_PMAC **pmac,
			const char *calling_fname )
{
	const char fname[] = "mxd_pmac_motor_get_pmac_pointer()";

	MX_RECORD *pmac_record;

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac == (MX_PMAC **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC pointer passed by '%s' was NULL.",
			calling_fname );
	}

	pmac_record = pmac_motor->pmac_record;

	if ( pmac_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The pmac_record pointer for record '%s' is NULL.",
			pmac_motor->record->name );
	}

	*pmac = (MX_PMAC *) pmac_record->record_type_struct;

	if ( *pmac == (MX_PMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC pointer for record '%s' is NULL.",
			pmac_motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_create_record_structures()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	pmac_motor = (MX_PMAC_MOTOR *) malloc( sizeof(MX_PMAC_MOTOR) );

	if ( pmac_motor == (MX_PMAC_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PMAC_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = pmac_motor;
	record->class_specific_function_list
				= &mxd_pmac_motor_function_list;

	motor->record = record;
	pmac_motor->record = record;

	/* A PMAC motor is treated as an stepper motor. */

	motor->subclass = MXC_MTR_STEPPER;

	/* We express accelerations in counts/sec**2. */

	motor->acceleration_type = MXF_MTR_ACCEL_RATE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_finish_record_initialization()";

	MX_PMAC_MOTOR *pmac_motor;
	double delay_seconds;
	mx_status_type mx_status;

	mx_status = mx_motor_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*** Server minimum delay time between commands. ***/

	delay_seconds = 0.0;

	record->event_interval
		= mx_convert_seconds_to_clock_ticks( delay_seconds );

	pmac_motor = (MX_PMAC_MOTOR *) record->record_type_struct;

	if ( ( pmac_motor->motor_number < 1 )
				|| ( pmac_motor->motor_number > 32 ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Illegal PMAC motor number %d for record '%s'.  Allowed range is 1 to 32.",
			pmac_motor->motor_number, record->name );
	}

	pmac_motor->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_print_structure()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	long motor_steps;
	double position, backlash, negative_limit, positive_limit, deadband;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	fprintf(file, "  Motor type     = PMAC motor.\n\n");

	fprintf(file, "  name           = %s\n", record->name);
	fprintf(file, "  port name      = %s\n", pmac_motor->pmac_record->name);
	fprintf(file, "  motor number   = %d\n", pmac_motor->motor_number);

	mx_status = mx_motor_get_position_steps( record, &motor_steps );

	if ( mx_status.code != MXE_SUCCESS ) {
		mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
				record->name );
	}
	
	position = motor->offset + motor->scale
			* (double) motor_steps;
	
	fprintf(file, "  position       = %ld steps (%g %s)\n",
			motor_steps, position, motor->units);
	fprintf(file, "  scale          = %g %s per step.\n",
			motor->scale, motor->units);
	fprintf(file, "  offset         = %g %s.\n",
			motor->offset, motor->units);
	
	backlash = motor->scale
			* (double) (motor->raw_backlash_correction.stepper);
	
	fprintf(file, "  backlash       = %ld steps (%g %s)\n",
			motor->raw_backlash_correction.stepper,
			backlash, motor->units);
	
	negative_limit = motor->offset
	+ motor->scale * (double)(motor->raw_negative_limit.stepper);
	
	fprintf(file, "  negative limit = %ld steps (%g %s)\n",
			motor->raw_negative_limit.stepper,
			negative_limit, motor->units);

	positive_limit = motor->offset
	+ motor->scale * (double)(motor->raw_positive_limit.stepper);

	fprintf(file, "  positive limit = %ld steps (%g %s)\n",
			motor->raw_positive_limit.stepper,
			positive_limit, motor->units);

	deadband = motor->scale * (double) motor->deadband.stepper;

	fprintf(file, "  deadband       = %ld steps (%g %s)\n\n",
			motor->deadband.stepper,
			deadband, motor->units );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_read_parms_from_hardware()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	double position;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Get the current position.  This has the side effect of
	 * updating the position value in the MX_MOTOR structure.
	 */

	mx_status = mx_motor_get_position( record, &position );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_resynchronize()";

	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	motor = (MX_MOTOR *) record->record_class_struct;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_resynchronize_record( pmac_motor->pmac_record );

	return mx_status;
}

/* ============ Motor specific functions ============ */

MX_EXPORT mx_status_type
mxd_pmac_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_motor_is_busy()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int length, open_loop_mode, in_position;
	unsigned long second_character, twelfth_character;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "?",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	second_character  = mx_hex_char_to_unsigned_long( response[1] );
	twelfth_character = mx_hex_char_to_unsigned_long( response[11] );

	MX_DEBUG( 2,("%s: second_character  = %#lx", fname, second_character));
	MX_DEBUG( 2,("%s: twelfth_character = %#lx", fname, twelfth_character));

	open_loop_mode        = (int)( second_character  & 0x4 );
	in_position           = (int)( twelfth_character & 0x1 );

	MX_DEBUG( 2,("%s: open_loop_mode = %d", fname, open_loop_mode ));
	MX_DEBUG( 2,("%s: in_position    = %d", fname, in_position));

	if ( open_loop_mode || in_position ) {
		motor->busy = FALSE;
	} else {
		motor->busy = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_move_absolute()";

	MX_PMAC_MOTOR *pmac_motor;
	MX_PMAC *pmac;
	char command[20];
	long motor_steps;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor_steps = motor->raw_destination.stepper;

	sprintf( command, "J=%ld", motor_steps );

	mx_status = mxd_pmac_jog_command( pmac_motor, command,
						NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_get_position_steps()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int num_tokens;
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "P",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The PMAC may report values back in terms of fractions of
	 * an encoder count.
	 */

	num_tokens = sscanf( response, "%lg", &double_value );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"No position value seen in response to 'P' command.  "
		"Response seen = '%s'", response );
	}

	/* We round this to the nearest encoder count. */

	motor->raw_position.stepper = mx_round( double_value );

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_pmac_configure_position_registers( MX_PMAC *pmac,
					MX_PMAC_MOTOR *pmac_motor )
{
	char command[40];
	char *ptr;
	int num;
	mx_status_type mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command, "@%x", pmac_motor->card_number );

		ptr = command + strlen( command );
	} else {
		strcpy( command, "" );

		ptr = command;
	}

	num = pmac_motor->motor_number;

	if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
		sprintf( ptr, "M%d64->D:$%XC", num, 4 + 8 * num );
	} else {
		sprintf( ptr, "M%d64->D:$%X", num, 192 * (num - 1) + 2067 );
	}

	mx_status = mxi_pmac_command( pmac, command, NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_set_position()";

	MX_PMAC_MOTOR *pmac_motor;
	MX_PMAC *pmac;
	long motor_position_scale_factor;
	long raw_difference, position_offset_change;
	char command[40];
	char response[40];
	char *ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Check to see if the Mx64 command is defined or not. */

	if ( pmac->num_cards > 1 ) {
		sprintf( command, "@%x", pmac_motor->card_number );

		ptr = command + strlen( command );
	} else {
		strcpy( command, "" );

		ptr = command;
	}

	sprintf( ptr, "M%d64->", pmac_motor->motor_number );

	mx_status = mxi_pmac_command( pmac, command,
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	ptr = response + strspn( response, " \t\r\n" );

	if ( *ptr == '*' ) {
		/* The Mx64 is not defined, so we attempt to define its
		 * recommended value here.
		 */

		mx_status = mxd_pmac_configure_position_registers(
							pmac, pmac_motor );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* Get the motor position scale factor.  The normal value is 96. */

	mx_status = mxd_pmac_get_motor_variable( pmac_motor, 8, MXFT_LONG,
						&motor_position_scale_factor,
						PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Get the old reported position in motor steps or encoder ticks. */

	mx_status = mx_motor_get_position( motor->record, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Compute the change to the reported position in motor steps
	 * or encoder ticks.
	 */

	raw_difference = motor->raw_set_position.stepper
				- motor->raw_position.stepper;

	/* Compute new position offset. */

	position_offset_change
		= 32 * raw_difference * motor_position_scale_factor;

	/* Change the position offset. */

	if ( pmac->num_cards > 1 ) {
		sprintf( command, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command, "" );
	}

	ptr = command + strlen( command );

	sprintf( ptr, "M%d64=M%d64+%ld",
			pmac_motor->motor_number,
			pmac_motor->motor_number,
			position_offset_change );

	mx_status = mxi_pmac_command( pmac, command,
				NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_soft_abort()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "J/",
						NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_immediate_abort()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "K",
						NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_positive_limit_hit()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int length;
	unsigned long first_character;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "?",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	first_character = mx_hex_char_to_unsigned_long( response[0] );

	MX_DEBUG( 2,("%s: first_character = %#lx", fname, first_character));

	if ( first_character & 0x2 ) {
		motor->positive_limit_hit = TRUE;
	} else {
		motor->positive_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_negative_limit_hit()";

	MX_PMAC_MOTOR *pmac_motor;
	char response[50];
	int length;
	unsigned long first_character;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "?",
				response, sizeof response, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	length = strlen( response );

	if ( length < 12 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Too few characters seen in response to motor status command.  "
	"Only saw %d characters.", length );
	}

	first_character = mx_hex_char_to_unsigned_long( response[0] );

	MX_DEBUG( 2,("%s: first_character = %#lx", fname, first_character));

	if ( first_character & 0x4 ) {
		motor->negative_limit_hit = TRUE;
	} else {
		motor->negative_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_find_home_position()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_pmac_jog_command( pmac_motor, "HM",
						NULL, 0, PMAC_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_constant_velocity_move()";

	MX_PMAC_MOTOR *pmac_motor;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( motor->constant_velocity_move > 0 ) {
		mx_status = mxd_pmac_jog_command( pmac_motor, "J+",
						NULL, 0, PMAC_DEBUG );
	} else
	if ( motor->constant_velocity_move < 0 ) {
		mx_status = mxd_pmac_jog_command( pmac_motor, "J-",
						NULL, 0, PMAC_DEBUG );
	} else {
		mx_status = MX_SUCCESSFUL_RESULT;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_get_parameter()";

	MX_PMAC_MOTOR *pmac_motor;
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		mx_status = mxd_pmac_get_motor_variable( pmac_motor, 22,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Reported by the controller in count/msec. */

		motor->raw_speed = 1000.0 * double_value;
		break;
	case MXLV_MTR_BASE_SPEED:
		/* The PMAC doesn't seem to have the concept of a base speed,
		 * since it is oriented to servo motors.
		 */

		motor->raw_base_speed = 0.0;

		break;
	case MXLV_MTR_ACCELERATION_PARAMETERS:
		mx_status = mxd_pmac_get_motor_variable( pmac_motor, 19,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		double_value = 1000000.0 * double_value;

		motor->acceleration_parameters[0] = double_value;

		motor->acceleration_parameters[1] = 0.0;
		motor->acceleration_parameters[2] = 0.0;
		motor->acceleration_parameters[3] = 0.0;
		break;
	default:
		return mx_motor_default_get_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_set_parameter()";

	MX_PMAC_MOTOR *pmac_motor;
	double double_value;
	mx_status_type mx_status;

	mx_status = mxd_pmac_get_pointers( motor, &pmac_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		double_value = 0.001 * motor->raw_speed;

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 22,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;
	case MXLV_MTR_BASE_SPEED:
		/* The PMAC doesn't seem to have the concept of a base speed,
		 * since it is oriented to servo motors.
		 */

		break;
	case MXLV_MTR_ACCELERATION_PARAMETERS:

		/* These acceleration parameters are chosen such that the
		 * value of Ixx19 is used as counts/msec**2.
		 */

		double_value = motor->acceleration_parameters[0];

		double_value = 1.0e-6 * double_value;

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 19,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Set Ixx20 (jog acceleration time) to 1.0 msec. */

		double_value = 1.0;

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 20,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Set Ixx21 (jog acceleration S-curve time) to 0. */

		double_value = 0.0;

		mx_status = mxd_pmac_set_motor_variable( pmac_motor, 21,
						MXFT_DOUBLE, &double_value,
						PMAC_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		break;

	default:
		return mx_motor_default_set_parameter_handler( motor );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_simultaneous_start( int num_motor_records,
				MX_RECORD **motor_record_array,
				double *position_array,
				int flags )
{
	const char fname[] = "mxd_pmac_simultaneous_start()";

	MX_RECORD *pmac_interface_record, *motor_record;
	MX_MOTOR *motor;
	MX_PMAC_MOTOR *pmac_motor;
	MX_PMAC *pmac;
	char command_buffer[500];
	char *ptr;
	int i;
	double raw_position;
	long raw_steps;
	mx_status_type mx_status;

	if ( num_motor_records <= 0 )
		return MX_SUCCESSFUL_RESULT;

	pmac_interface_record = NULL;
	pmac = NULL;

	/* Construct the move command to send to the PMAC. */

	ptr = command_buffer;

	*ptr = '\0';

	for ( i = 0; i < num_motor_records; i++ ) {
		motor_record = motor_record_array[i];

		motor = (MX_MOTOR *) motor_record->record_class_struct;

		if ( motor_record->mx_type != MXT_MTR_PMAC ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
			"Cannot perform a simultaneous start since motors "
			"'%s' and '%s' are not the same type of motors.",
				motor_record_array[0]->name,
				motor_record->name );
		}

		pmac_motor = (MX_PMAC_MOTOR *) motor_record->record_type_struct;

		if ( pmac_interface_record == (MX_RECORD *) NULL ) {
			pmac_interface_record = pmac_motor->pmac_record;

			pmac = (MX_PMAC *)
				pmac_interface_record->record_type_struct;
		}

		/* Verify that the PMAC motor records all belong to the same
		 * PMAC interface.
		 */

		if ( pmac_interface_record != pmac_motor->pmac_record ) {
			return mx_error( MXE_UNSUPPORTED, fname,
		"Cannot perform a simultaneous start for motors '%s' and '%s' "
		"since they are controlled by different PMAC interfaces, "
		"namely '%s' and '%s'.",
				motor_record_array[0]->name,
				motor_record->name,
				pmac_interface_record->name,
				pmac_motor->pmac_record->name );
		}

		/* Compute the new position in raw units. */

		raw_position =
			mx_divide_safely( position_array[i] - motor->offset,
						motor->scale );

		raw_steps = mx_round( raw_position );

		/* Append the part of the command referring to this motor. */

		ptr = command_buffer + strlen( command_buffer );

		if ( pmac->num_cards > 1 ) {
			sprintf( ptr, "@%x", pmac_motor->card_number );
		}

		ptr = command_buffer + strlen( command_buffer );

		sprintf(ptr, "#%dJ=%ld ", pmac_motor->motor_number, raw_steps);
	}

	if ( pmac_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"No pmac interface record pointer was found for record '%s'.",
		motor_record_array[i]->name );
	}

	/* Send the command to the PMAC. */

	MX_DEBUG( 2,("%s: command = '%s'", fname, command_buffer));

	mx_status = mxi_pmac_command( pmac, command_buffer,
						NULL, 0, PMAC_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_jog_command( MX_PMAC_MOTOR *pmac_motor,
			char *command,
			char *response,
			int response_buffer_length,
			int debug_flag )
{
	const char fname[] = "mxd_pmac_jog_command()";

	MX_PMAC *pmac;
	char command_buffer[100];
	char *ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command_buffer, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command_buffer, "" );
	}

	ptr = command_buffer + strlen( command_buffer );

	sprintf( ptr, "#%d%s", pmac_motor->motor_number, command );

	mx_status = mxi_pmac_command( pmac, command_buffer,
				response, response_buffer_length, debug_flag );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_get_motor_variable( MX_PMAC_MOTOR *pmac_motor,
				int variable_number,
				int variable_type,
				void *variable_ptr,
				int debug_flag )
{
	const char fname[] = "mxd_pmac_get_motor_variable()";

	MX_PMAC *pmac;
	char command_buffer[100];
	char response[100];
	char *ptr;
	int num_items;
	long long_value;
	long *long_ptr;
	double double_value;
	double *double_ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command_buffer, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command_buffer, "" );
	}

	ptr = command_buffer + strlen( command_buffer );

	if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
		sprintf( ptr, "I%02d%02d",
				pmac_motor->motor_number, variable_number );
	} else {
		sprintf( ptr, "I%d%02d",
				pmac_motor->motor_number, variable_number );
	}

	mx_status = mxi_pmac_command( pmac, command_buffer,
				response, sizeof response, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( variable_type ) {
	case MXFT_LONG:
		num_items = sscanf( response, "%ld", &long_value );

		if ( num_items != 1 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
"Cannot parse response to command '%s' as a long integer.  Response = '%s'",
				command_buffer, response );
		}
		long_ptr = (long *) variable_ptr;

		*long_ptr = long_value;
		break;
	case MXFT_DOUBLE:
		num_items = sscanf( response, "%lg", &double_value );

		if ( num_items != 1 ) {
			return mx_error( MXE_UNPARSEABLE_STRING, fname,
"Cannot parse response to command '%s' as a double.  Response = '%s'",
				command_buffer, response );
		}
		double_ptr = (double *) variable_ptr;

		*double_ptr = double_value;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Only MXFT_LONG and MXFT_DOUBLE are supported." );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_set_motor_variable( MX_PMAC_MOTOR *pmac_motor,
				int variable_number,
				int variable_type,
				void *variable_ptr,
				int debug_flag )
{
	const char fname[] = "mxd_pmac_set_motor_variable()";

	MX_PMAC *pmac;
	char command_buffer[100];
	char response[100];
	char *ptr;
	long *long_ptr;
	double *double_ptr;
	mx_status_type mx_status;

	mx_status = mxd_pmac_motor_get_pmac_pointer( pmac_motor, &pmac, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( pmac->num_cards > 1 ) {
		sprintf( command_buffer, "@%x", pmac_motor->card_number );
	} else {
		strcpy( command_buffer, "" );
	}

	ptr = command_buffer + strlen( command_buffer );

	switch( variable_type ) {
	case MXFT_LONG:
		long_ptr = (long *) variable_ptr;

		if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
			sprintf( ptr, "I%02d%02d=%ld",
				pmac_motor->motor_number,
				variable_number,
				*long_ptr );
		} else {
			sprintf( ptr, "I%d%02d=%ld",
				pmac_motor->motor_number,
				variable_number,
				*long_ptr );
		}
		break;
	case MXFT_DOUBLE:
		double_ptr = (double *) variable_ptr;

		if ( pmac->pmac_type & MX_PMAC_TYPE_TURBO ) {
			sprintf( ptr, "I%02d%02d=%g",
				pmac_motor->motor_number,
				variable_number,
				*double_ptr );
		} else {
			sprintf( ptr, "I%d%02d=%g",
				pmac_motor->motor_number,
				variable_number,
				*double_ptr );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Only MXFT_LONG and MXFT_DOUBLE are supported." );
		break;
	}

	mx_status = mxi_pmac_command( pmac, command_buffer,
				response, sizeof response, debug_flag );

	return mx_status;
}

