/*
 * Name:    d_slit_motor.c 
 *
 * Purpose: MX motor driver to use a pair of MX motors as the sides
 *          of a slit.  Two records are needed to completely control
 *          a set of slit motors.  One record is for the slit center
 *          and one record is for the slit width.
 *
 * Note:    This driver may be used both for slit blade pairs for which
 *          positive moves are in the same direction and for slit blade
 *          pairs for which positive moves are in opposite directions.
 *          See 'd_slit_motor.h' for more information.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_motor.h"
#include "d_slit_motor.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_slit_motor_record_function_list = {
	mxd_slit_motor_initialize_type,
	mxd_slit_motor_create_record_structures,
	mxd_slit_motor_finish_record_initialization,
	mxd_slit_motor_delete_record,
	mxd_slit_motor_print_motor_structure,
	mxd_slit_motor_read_parms_from_hardware,
	mxd_slit_motor_write_parms_to_hardware,
	mxd_slit_motor_open,
	mxd_slit_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_slit_motor_motor_function_list = {
	mxd_slit_motor_motor_is_busy,
	mxd_slit_motor_move_absolute,
	mxd_slit_motor_get_position,
	mxd_slit_motor_set_position,
	mxd_slit_motor_soft_abort,
	mxd_slit_motor_immediate_abort,
	mxd_slit_motor_positive_limit_hit,
	mxd_slit_motor_negative_limit_hit,
	mxd_slit_motor_find_home_position
};

/* Slit motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_slit_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_SLIT_MOTOR_STANDARD_FIELDS
};

long mxd_slit_motor_num_record_fields
		= sizeof( mxd_slit_motor_record_field_defaults )
			/ sizeof( mxd_slit_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_slit_motor_rfield_def_ptr
			= &mxd_slit_motor_record_field_defaults[0];

/* === */

MX_EXPORT mx_status_type
mxd_slit_motor_initialize_type( long type )
{
		/* Nothing needed here. */

		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_slit_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_SLIT_MOTOR *slit_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	slit_motor = (MX_SLIT_MOTOR *) malloc( sizeof(MX_SLIT_MOTOR) );

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SLIT_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = slit_motor;
	record->class_specific_function_list
				= &mxd_slit_motor_motor_function_list;

	motor->record = record;

	/* A slit motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_finish_record_initialization( MX_RECORD *record )
{
	MX_MOTOR *motor;

	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor = (MX_MOTOR *) record->record_class_struct;

	motor->motor_flags |= MXF_MTR_IS_PSEUDOMOTOR;
	motor->motor_flags |= MXF_MTR_CANNOT_QUICK_SCAN;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_print_motor_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_slit_motor_print_motor_structure()";

	MX_MOTOR *motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	MX_SLIT_MOTOR *slit_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	slit_motor = (MX_SLIT_MOTOR *) (record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	positive_motor_record = slit_motor->positive_motor_record;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type      = SLIT_MOTOR.\n\n");

	fprintf(file, "  name            = %s\n", record->name);
	fprintf(file, "  slit type       = %ld\n", slit_motor->slit_type);
	fprintf(file, "  negative motor  = %s\n",
					negative_motor_record->name);
	fprintf(file, "  positive motor  = %s\n",
					positive_motor_record->name);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position        = %g %s\n",
		position, motor->units);
	fprintf(file, "  slit scale    = %g %s per unscaled slit unit.\n",
		motor->scale, motor->units);
	fprintf(file, "  slit offset   = %g %s.\n",
		motor->offset, motor->units);
	fprintf(file, "  backlash        = %g %s.\n",
		motor->raw_backlash_correction.analog, motor->units);
	fprintf(file, "  negative limit  = %g %s.\n",
		motor->raw_negative_limit.analog, motor->units);
	fprintf(file, "  positive limit  = %g %s.\n",
		motor->raw_positive_limit.analog, motor->units);

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband        = %g %s.\n\n",
		deadband, motor->units );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by the MX_MOTOR records
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by the MX_MOTOR records
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_open( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_close( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_motor_is_busy()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	int busy;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_is_busy( negative_motor_record, &busy );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* If the negative motor is busy, we are done. */

	if ( busy == TRUE ) {
		motor->busy = TRUE;

		return MX_SUCCESSFUL_RESULT;
	}

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_is_busy( positive_motor_record, &busy );

	motor->busy = busy;

	return status;
}

MX_EXPORT mx_status_type
mxd_slit_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_move_absolute()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	MX_RECORD *motor_record_array[2];
	double new_slit_position, old_slit_position;
	double new_negative_motor_position, old_negative_motor_position;
	double new_positive_motor_position, old_positive_motor_position;
	double new_position_array[2];
	double slit_position_difference;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	new_slit_position = motor->raw_destination.analog;

	/* Check the limits on this request. */

	if ( new_slit_position > motor->raw_positive_limit.analog
	  || new_slit_position < motor->raw_negative_limit.analog )
	{
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Requested position %g is outside the allowed range of %g to %g",
			new_slit_position, motor->raw_negative_limit.analog,
			motor->raw_positive_limit.analog );
	}

	/* Get the old motor and slit positions. */

	status = mx_motor_get_position(
			negative_motor_record, &old_negative_motor_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_motor_get_position(
			positive_motor_record, &old_positive_motor_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the new motor positions. */

	switch( slit_motor->slit_type ) {
	case MXF_SLIT_CENTER_SAME:
		old_slit_position = 0.5 *
		  (old_positive_motor_position + old_negative_motor_position);

		slit_position_difference
			= new_slit_position - old_slit_position;

		new_negative_motor_position = old_negative_motor_position
					+ slit_position_difference;

		new_positive_motor_position = old_positive_motor_position
					+ slit_position_difference;
		break;
	case MXF_SLIT_WIDTH_SAME:
		old_slit_position = old_positive_motor_position
					- old_negative_motor_position;

		slit_position_difference
			= new_slit_position - old_slit_position;

		new_negative_motor_position = old_negative_motor_position
					- 0.5 * slit_position_difference;

		new_positive_motor_position = old_positive_motor_position
					+ 0.5 * slit_position_difference;
		break;
	case MXF_SLIT_CENTER_OPPOSITE:
		old_slit_position = 0.5 *
		  (old_positive_motor_position - old_negative_motor_position);

		slit_position_difference
			= new_slit_position - old_slit_position;

		new_negative_motor_position = old_negative_motor_position
					- slit_position_difference;

		new_positive_motor_position = old_positive_motor_position
					+ slit_position_difference;
		break;
	case MXF_SLIT_WIDTH_OPPOSITE:
		old_slit_position = old_positive_motor_position
					+ old_negative_motor_position;

		slit_position_difference
			= new_slit_position - old_slit_position;

		new_negative_motor_position = old_negative_motor_position
					+ 0.5 * slit_position_difference;

		new_positive_motor_position = old_positive_motor_position
					+ 0.5 * slit_position_difference;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown slit motor type %ld for slit motor '%s'",
			slit_motor->slit_type, motor->record->name );
		break;
	}

	/* Now perform the move. */

	motor_record_array[0] = negative_motor_record;
	motor_record_array[1] = positive_motor_record;

	new_position_array[0] = new_negative_motor_position;
	new_position_array[1] = new_positive_motor_position;

	status = mx_motor_array_move_absolute( 2, motor_record_array,
			new_position_array, MXF_MTR_NOWAIT );

	return status;
}

MX_EXPORT mx_status_type
mxd_slit_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_get_position()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	double slit_position;
	double negative_motor_position, positive_motor_position;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	/* Get the current motor positions. */

	status = mx_motor_get_position(
			negative_motor_record, &negative_motor_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_motor_get_position(
			positive_motor_record, &positive_motor_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the new motor positions. */

	switch( slit_motor->slit_type ) {
	case MXF_SLIT_CENTER_SAME:
		slit_position = 0.5 *
			(positive_motor_position + negative_motor_position);
		break;
	case MXF_SLIT_WIDTH_SAME:
		slit_position = positive_motor_position
					- negative_motor_position;
		break;
	case MXF_SLIT_CENTER_OPPOSITE:
		slit_position = 0.5 *
			(positive_motor_position - negative_motor_position);
		break;
	case MXF_SLIT_WIDTH_OPPOSITE:
		slit_position = positive_motor_position
					+ negative_motor_position;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown slit motor type %ld for slit motor '%s'",
			slit_motor->slit_type, motor->record->name );
		break;
	}

	motor->raw_position.analog = slit_position;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_slit_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_set_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'set position' is not supported for a slit pseudomotor." );
}

MX_EXPORT mx_status_type
mxd_slit_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_soft_abort()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_soft_abort( negative_motor_record );

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_soft_abort( positive_motor_record );

	return status;
}

MX_EXPORT mx_status_type
mxd_slit_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_immediate_abort()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_immediate_abort( negative_motor_record );

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_immediate_abort( positive_motor_record );

	return status;
}

MX_EXPORT mx_status_type
mxd_slit_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_positive_limit_hit()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	int limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	limit_hit = FALSE;

	switch( slit_motor->slit_type ) {
	case MXF_SLIT_CENTER_SAME:
	case MXF_SLIT_WIDTH_OPPOSITE:
		if ( motor->scale >= 0.0 ) {
			status = mx_motor_positive_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_positive_limit_hit(
					negative_motor_record, &limit_hit );
		} else {
			status = mx_motor_negative_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_negative_limit_hit(
					negative_motor_record, &limit_hit );
		}
		break;
	case MXF_SLIT_WIDTH_SAME:
	case MXF_SLIT_CENTER_OPPOSITE:
		if ( motor->scale >= 0.0 ) {
			status = mx_motor_positive_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_negative_limit_hit(
					negative_motor_record, &limit_hit );
		} else {
			status = mx_motor_negative_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_positive_limit_hit(
					negative_motor_record, &limit_hit );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal slit type %ld for slit motor '%s'",
			slit_motor->slit_type, motor->record->name );
		break;
	}

	motor->positive_limit_hit = limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mxd_slit_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_negative_limit_hit()";

	MX_SLIT_MOTOR *slit_motor;
	MX_RECORD *negative_motor_record, *positive_motor_record;
	int limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	slit_motor = (MX_SLIT_MOTOR *) (motor->record->record_type_struct);

	if ( slit_motor == (MX_SLIT_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SLIT_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	positive_motor_record = slit_motor->positive_motor_record;

	if ( positive_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	negative_motor_record = slit_motor->negative_motor_record;

	if ( negative_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	limit_hit = FALSE;

	switch( slit_motor->slit_type ) {
	case MXF_SLIT_CENTER_SAME:
	case MXF_SLIT_WIDTH_OPPOSITE:
		if ( motor->scale >= 0.0 ) {
			status = mx_motor_negative_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_negative_limit_hit(
					negative_motor_record, &limit_hit );
		} else {
			status = mx_motor_positive_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_positive_limit_hit(
					negative_motor_record, &limit_hit );
		}
		break;
	case MXF_SLIT_WIDTH_SAME:
	case MXF_SLIT_CENTER_OPPOSITE:
		if ( motor->scale >= 0.0 ) {
			status = mx_motor_negative_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_positive_limit_hit(
					negative_motor_record, &limit_hit );
		} else {
			status = mx_motor_positive_limit_hit(
					positive_motor_record, &limit_hit );

			if ( limit_hit == TRUE )
				break;

			status = mx_motor_negative_limit_hit(
					negative_motor_record, &limit_hit );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal slit type %ld for slit motor '%s'",
			slit_motor->slit_type, motor->record->name );
		break;
	}

	motor->negative_limit_hit = limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mxd_slit_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_slit_motor_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"Finding home for a slit motor is _not_ allowed.  Motor name = '%s'",
		motor->record->name );
}

