/*
 * Name:    d_xia_dxp_mca.c
 *
 * Purpose: MX multichannel analyzer driver for the X-Ray Instrumentation
 *          Associates Mesa2x Data Server.  This driver controls an individual
 *          MCA in a multi-MCA module.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "mxconfig.h"
#include "mx_osdef.h"

#if USE_TCPIP || HAVE_XIA_XERXES

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_types.h"
#include "mx_socket.h"
#include "mx_mca.h"

#if HAVE_XIA_XERXES

#include <xerxes_errors.h>
#include <xerxes_structures.h>
#include <xerxes.h>

#include "i_xia_xerxes.h"
#endif

#include "i_xia_mds.h"
#include "i_xia_network.h"
#include "d_xia_dxp_mca.h"

/* Initialize the MCA driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_xia_dxp_record_function_list = {
	mxd_xia_dxp_initialize_type,
	mxd_xia_dxp_create_record_structures,
	mxd_xia_dxp_finish_record_initialization,
	mxd_xia_dxp_delete_record,
	mxd_xia_dxp_print_structure,
	mxd_xia_dxp_read_parms_from_hardware,
	mxd_xia_dxp_write_parms_to_hardware,
	mxd_xia_dxp_open,
	mxd_xia_dxp_close,
	NULL,
	NULL,
	mxd_xia_dxp_special_processing_setup
};

MX_MCA_FUNCTION_LIST mxd_xia_dxp_mca_function_list = {
	mxd_xia_dxp_start,
	mxd_xia_dxp_stop,
	mxd_xia_dxp_read,
	mxd_xia_dxp_clear,
	mxd_xia_dxp_busy,
	mxd_xia_dxp_get_parameter,
	mxd_xia_dxp_set_parameter,
};

MX_RECORD_FIELD_DEFAULTS mxd_xia_dxp_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCA_STANDARD_FIELDS,
	MXD_XIA_DXP_STANDARD_FIELDS
};

long mxd_xia_dxp_num_record_fields
		= sizeof( mxd_xia_dxp_record_field_defaults )
		  / sizeof( mxd_xia_dxp_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_xia_dxp_rfield_def_ptr
			= &mxd_xia_dxp_record_field_defaults[0];

#define MXD_XIA_DXP_DEBUG	TRUE

#define MXD_XIA_VERIFY_PRESETS	TRUE

static mx_status_type mxd_xia_dxp_process_function( void *record_ptr,
						void *record_field_ptr,
						int operation );

/* A private function for the use of the driver. */

static mx_status_type
mxd_xia_dxp_get_pointers( MX_MCA *mca,
			MX_XIA_DXP_MCA **xia_dxp_mca,
			const char *calling_fname )
{
	const char fname[] = "mxd_xia_dxp_get_pointers()";

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_dxp_mca == (MX_XIA_DXP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_DXP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_MCA pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*xia_dxp_mca = (MX_XIA_DXP_MCA *) mca->record->record_type_struct;

	if ( (*xia_dxp_mca)->xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The xia_dxp_record pointer for record '%s' is NULL.",
			mca->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_xia_dxp_initialize_type( long record_type )
{
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long num_record_fields;
	long maximum_num_channels_varargs_cookie;
	long maximum_num_rois_varargs_cookie;
	long num_soft_rois_varargs_cookie;
	mx_status_type mx_status;

	mx_status = mx_mca_initialize_type( record_type,
				&num_record_fields,
				&record_field_defaults,
				&maximum_num_channels_varargs_cookie,
				&maximum_num_rois_varargs_cookie,
				&num_soft_rois_varargs_cookie );

	return mx_status;
}


MX_EXPORT mx_status_type
mxd_xia_dxp_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_xia_dxp_create_record_structures()";

	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;

	/* Allocate memory for the necessary structures. */

	mca = (MX_MCA *) malloc( sizeof(MX_MCA) );

	if ( mca == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCA structure." );
	}

	xia_dxp_mca = (MX_XIA_DXP_MCA *) malloc( sizeof(MX_XIA_DXP_MCA) );

	if ( xia_dxp_mca == (MX_XIA_DXP_MCA *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_XIA_DXP_MCA structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mca;
	record->record_type_struct = xia_dxp_mca;
	record->class_specific_function_list = &mxd_xia_dxp_mca_function_list;

	mca->record = record;
	xia_dxp_mca->record = record;

	xia_dxp_mca->is_busy = NULL;
	xia_dxp_mca->read_parameter = NULL;
	xia_dxp_mca->write_parameter = NULL;
	xia_dxp_mca->write_parameter_to_all_channels = NULL;
	xia_dxp_mca->start_run = NULL;
	xia_dxp_mca->stop_run = NULL;
	xia_dxp_mca->read_spectrum = NULL;
	xia_dxp_mca->read_statistics = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_xia_dxp_finish_record_initialization()";

	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed is NULL." );
	}

	mca = (MX_MCA *) record->record_class_struct;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCA pointer for record '%s' is NULL.", record->name );
	}

	if ( mca->maximum_num_channels > 8192 ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"The requested maximum number of channels (%lu) for XIA MCA '%s' is greater "
"than the maximum allowed value of 8192.",
			mca->maximum_num_channels, record->name );
	}

	mca->channel_array = ( unsigned long * )
		malloc( mca->maximum_num_channels * sizeof( unsigned long ) );

	if ( mca->channel_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory allocating an %lu channel data array.",
			mca->maximum_num_channels );
	}

	mca->preset_type = MXF_MCA_PRESET_LIVE_TIME;

	xia_dxp_mca = (MX_XIA_DXP_MCA *) record->record_type_struct;

	xia_dxp_mca->new_statistics_available = TRUE;

	mx_status = mx_mca_finish_record_initialization( record );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_xia_dxp_print_structure()";

	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	mca = (MX_MCA *) (record->record_class_struct);

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fprintf(file, "MCA parameters for record '%s':\n", record->name);

	fprintf(file, "  MCA type              = XIA_DXP_MCA.\n\n");
	fprintf(file, "  XIA DXP record        = '%s'\n",
					xia_dxp_mca->xia_dxp_record->name);
	fprintf(file, "  mca label             = %s\n",
					xia_dxp_mca->mca_label);
	fprintf(file, "  maximum # of bins     = %lu\n",
					mca->maximum_num_channels);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

/* The following routine does Mesa Data Server specific initialization. */

static mx_status_type
mxd_xia_dxp_mds_open( MX_MCA *mca,
			MX_XIA_DXP_MCA *xia_dxp_mca,
			MX_RECORD *xia_dxp_record )
{
	const char fname[] = "mxd_xia_dxp_mds_open()";

	MX_XIA_MDS *xia_mds;
	unsigned long *config_response;
	unsigned long i, j;
	int current_crate, current_slot, current_dxp_channel;
	int previous_crate, previous_slot, dxp_module_number;

	xia_dxp_mca->system_channel = atoi( xia_dxp_mca->mca_label );

	xia_mds = (MX_XIA_MDS *) xia_dxp_record->record_type_struct;

	if ( xia_mds->num_mcas <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record '%s' used by MCA '%s' has not yet been configured.  "
	"You should verify that the entry for MCA '%s' is listed in the "
	"MX database file _after_ the entry for '%s'.",
			xia_dxp_record->name, mca->record->name,
			mca->record->name, xia_dxp_record->name );
	}

#if HAVE_XIA_XERXES
	xia_dxp_mca->use_mca_channel_array = TRUE;
#endif

	/* Set up the interface function pointers. */

	xia_dxp_mca->is_busy = mxi_xia_mds_is_busy;
	xia_dxp_mca->read_parameter = mxi_xia_mds_read_parameter;
	xia_dxp_mca->write_parameter = mxi_xia_mds_write_parameter;
	xia_dxp_mca->write_parameter_to_all_channels
				= mxi_xia_mds_write_parameter_to_all_channels;
	xia_dxp_mca->start_run = mxi_xia_mds_start_run;
	xia_dxp_mca->stop_run = mxi_xia_mds_stop_run;
	xia_dxp_mca->read_spectrum = mxi_xia_mds_read_spectrum;
	xia_dxp_mca->read_statistics = mxi_xia_mds_read_statistics;

	/* Walk through the list of MCAs returned by 'Restore Configuration
	 * From File' looking for this MCA.
	 */

	config_response = xia_mds->restore_configuration_response;

	dxp_module_number = 0;

	previous_crate = (int) config_response[0];
	previous_slot  = (int) config_response[1];

	for ( i = 0; i < xia_mds->num_mcas; i++ ) {

		j = 3 * i;

		current_crate       = (int) config_response[j];
		current_slot        = (int) config_response[j+1];
		current_dxp_channel = (int) config_response[j+2];

		/* Not all DXP channels in a DXP module are necessarily
		 * in use.  This prevents a simple calculation of the
		 * module number.
		 */

		if ( ( current_crate != previous_crate )
		  || ( current_slot  != previous_slot  ) )
		{
			dxp_module_number++;
		}

		if ( i == xia_dxp_mca->system_channel ) {

			if ( xia_mds->mca_record_array[i] == (MX_RECORD *)NULL )
			{
				xia_mds->mca_record_array[i] = mca->record;

				xia_dxp_mca->dxp_module  = dxp_module_number;
				xia_dxp_mca->crate       = current_crate;
				xia_dxp_mca->slot        = current_slot;
				xia_dxp_mca->dxp_channel = current_dxp_channel;

				MX_DEBUG( 2,
			("%s: '%s' (%d, %d, %d) -> "
			"'%s' mca_record_array[%lu], module %d",
					fname, mca->record->name,
					xia_dxp_mca->crate,
					xia_dxp_mca->slot,
					xia_dxp_mca->dxp_channel,
					xia_dxp_record->name, i,
					xia_dxp_mca->dxp_module ));

				break;	/* Exit the for() loop. */
			} else {
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"MCA '%s' (crate %d, slot %d, mca %d) should be saved in "
		"mca_record_array[%lu] for MCA controller '%s', but that "
		"location is already in use by record '%s'.",
					mca->record->name,
					xia_dxp_mca->crate,
					xia_dxp_mca->slot,
					xia_dxp_mca->dxp_channel,
					i, xia_dxp_record->name,
					xia_mds->mca_record_array[i]->name );
			}
		}

		previous_crate = current_crate;
		previous_slot  = current_slot;
	}
	if ( i >= xia_mds->num_mcas ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Error setting up MCA '%s' (crate %d, slot %d, mca %d).  "
		"MCA controller '%s' says that it does not know of an MCA "
		"at that address.", mca->record->name,
					xia_dxp_mca->crate,
					xia_dxp_mca->slot,
					xia_dxp_mca->dxp_channel,
					xia_dxp_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* The following routine does XIA network record specific initialization. */

static mx_status_type
mxd_xia_dxp_network_open( MX_MCA *mca,
			MX_XIA_DXP_MCA *xia_dxp_mca,
			MX_RECORD *xia_dxp_record )
{
	const char fname[] = "mxd_xia_dxp_network_open()";

	MX_XIA_NETWORK *xia_network;
	unsigned long i;

	xia_network = (MX_XIA_NETWORK *) xia_dxp_record->record_type_struct;

	if ( xia_network->num_mcas <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record '%s' used by MCA '%s' has not yet been configured.  "
	"You should verify that the entry for MCA '%s' is listed in the "
	"MX database file _after_ the entry for '%s'.",
			xia_dxp_record->name, mca->record->name,
			mca->record->name, xia_dxp_record->name );
	}

#if HAVE_XIA_XERXES
	xia_dxp_mca->use_mca_channel_array = TRUE;
#endif

	/* Set up the interface function pointers. */

	xia_dxp_mca->is_busy = mxi_xia_network_is_busy;
	xia_dxp_mca->read_parameter = mxi_xia_network_read_parameter;
	xia_dxp_mca->write_parameter = mxi_xia_network_write_parameter;
	xia_dxp_mca->write_parameter_to_all_channels
				= mxi_xia_network_write_param_to_all_channels;
	xia_dxp_mca->start_run = mxi_xia_network_start_run;
	xia_dxp_mca->stop_run = mxi_xia_network_stop_run;
	xia_dxp_mca->read_spectrum = mxi_xia_network_read_spectrum;
	xia_dxp_mca->read_statistics = mxi_xia_network_read_statistics;

	/* It does not matter which slot in mca_record_array that this
	 * record is put into, so pick the first empty one.
	 */

	for ( i = 0; i < xia_network->num_mcas; i++ ) {

		if ( xia_network->mca_record_array[i] == NULL ) {

			xia_network->mca_record_array[i] = mca->record;

			xia_dxp_mca->system_channel = (int) i;

			break;		/* Exit the for() loop. */
		}
	}

	if ( i >= xia_network->num_mcas ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Error adding MCA '%s' to the list of MCAs controlled by "
		"XIA network record '%s'.  This MCA would be record %lu "
		"of the MCA record array, but that array only has room "
		"for %lu records.", mca->record->name,
			xia_dxp_record->name, i+1, xia_network->num_mcas );
	}

	return MX_SUCCESSFUL_RESULT;
}

#if HAVE_XIA_XERXES

/* The following routine does XerXes specific initialization. */

static mx_status_type
mxd_xia_dxp_xerxes_open( MX_MCA *mca,
			MX_XIA_DXP_MCA *xia_dxp_mca,
			MX_RECORD *xia_dxp_record )
{
	const char fname[] = "mxd_xia_dxp_xerxes_open()";

	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	unsigned long i;

	xia_dxp_mca->system_channel = atoi( xia_dxp_mca->mca_label );

	xia_xerxes = (MX_XIA_XERXES *) xia_dxp_record->record_type_struct;

	if ( xia_xerxes->num_mcas <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record '%s' used by MCA '%s' has not yet been configured.  "
	"You should verify that the entry for MCA '%s' is listed in the "
	"MX database file _after_ the entry for '%s'.",
			xia_dxp_record->name, mca->record->name,
			mca->record->name, xia_dxp_record->name );
	}

	/* Set up the interface function pointers. */

	xia_dxp_mca->is_busy = mxi_xia_xerxes_is_busy;
	xia_dxp_mca->read_parameter = mxi_xia_xerxes_read_parameter;
	xia_dxp_mca->write_parameter = mxi_xia_xerxes_write_parameter;
	xia_dxp_mca->write_parameter_to_all_channels
			= mxi_xia_xerxes_write_parameter_to_all_channels;
	xia_dxp_mca->start_run = mxi_xia_xerxes_start_run;
	xia_dxp_mca->stop_run = mxi_xia_xerxes_stop_run;
	xia_dxp_mca->read_spectrum = mxi_xia_xerxes_read_spectrum;
	xia_dxp_mca->read_statistics = mxi_xia_xerxes_read_statistics;

	i = xia_dxp_mca->system_channel;

	if ( i >= xia_xerxes->num_mcas ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Error setting up MCA '%s' for system channel %d.  "
		"MCA controller '%s' says that the maximum system channel "
		"number is %ld.",
			mca->record->name,
			xia_dxp_mca->system_channel,
			xia_dxp_record->name,
			xia_xerxes->num_mcas );
	}

	if ( xia_xerxes->mca_record_array[i] == (MX_RECORD *) NULL ) {

		xia_xerxes->mca_record_array[i] = mca->record;

		xia_dxp_mca->dxp_module = -1;
		xia_dxp_mca->crate = -1;
		xia_dxp_mca->slot = -1;
		xia_dxp_mca->dxp_channel = -1;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"MCA '%s' (system channel %d) should be saved in "
		"mca_record_array[%lu] for MCA controller '%s', but that "
		"location is already in use by record '%s'.",
			xia_dxp_mca->record->name,
			xia_dxp_mca->system_channel, i,
			xia_dxp_record->name,
			xia_xerxes->mca_record_array[i]->name );
	}

	/* See how many bins are in the spectrum. */

	xia_status = dxp_nspec( &(xia_dxp_mca->system_channel),
					&(xia_dxp_mca->num_spectrum_bins) );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to get the number of spectrum bins for MCA '%s'.  "
		"Error code = %d, error status = '%s'",
			mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

#if 0
	/* Currently, it is an error for the number of bins in the spectrum
	 * to be different than mca->maximum_num_channels.  This policy
	 * was recommended by Jeff Terry of MR-CAT (APS), since the
	 * XIA system needs a different configuration file if there is a 
	 * different number of channels.
	 */

	if ( xia_dxp_mca->num_spectrum_bins != mca->maximum_num_channels ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The number of spectrum bins (%u) reported by the XIA MCA '%s' is different "
"from the maximum number of channels (%lu) configured for this MCA in the "
"MX database.  You must either reconfigure MX to match the number of "
"channels reported by XIA or load a XIA configuration that is configured "
"for %lu channels.", xia_dxp_mca->num_spectrum_bins,
			mca->maximum_num_channels,
			mca->maximum_num_channels );
	}
#endif

	/* Do we need to allocate memory for a spectrum array or is
	 * mca->channel_array already the right size?
	 */

	if ( ( sizeof(mx_uint32_type) == sizeof(unsigned long) )
	  && ( mca->maximum_num_channels >= xia_dxp_mca->num_spectrum_bins ) )
	{
		/* We do not need to allocate a spectrum array since
		 * mca->channel_array is already big enough and of the
		 * right data type to hold the data.  This allows us
		 * to skip the step of copying the data from the array
		 * xia_dxp_mca->spectrum_array to mca->channel_array.
		 */

		xia_dxp_mca->use_mca_channel_array = TRUE;

		xia_dxp_mca->spectrum_array = NULL;
	} else {
		/* We _do_ need to allocate a separate spectrum array.
		 * This will happen for one of two reasons:
		 *
		 * 1.  Either the value xia_dxp_mca->num_spectrum_bins
		 *     is bigger than mca->maximum_num_channels which
		 *     means that the spectrum returned by XIA would
		 *     not fit into mca->channel_array.
		 *
		 * 2.  We are running on a computer where unsigned longs
		 *     are not 32 bits, such as on an Alpha.
		 *
		 * Either way, we have to allocate a separate array
		 * and put up with the performance hit of copying the
		 * array each time.
		 */

		xia_dxp_mca->use_mca_channel_array = FALSE;

		xia_dxp_mca->spectrum_array = (unsigned long *) malloc(
		    xia_dxp_mca->num_spectrum_bins * sizeof(unsigned long) );

		if ( xia_dxp_mca->spectrum_array == (unsigned long *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocated a DXP spectrum array of %u unsigned longs",
			xia_dxp_mca->num_spectrum_bins );
		}
	}

	/* Allocate memory for the baseline array. */

	xia_status = dxp_nbase( &(xia_dxp_mca->system_channel),
					&(xia_dxp_mca->num_baseline_bins) );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to get the number of baseline bins for MCA '%s'.  "
		"Error code = %d, error status = '%s'",
			mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	xia_dxp_mca->baseline_array = (unsigned short *)
	    malloc( xia_dxp_mca->num_baseline_bins * sizeof(unsigned short) );

	if ( xia_dxp_mca->baseline_array == (unsigned short *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocated a DXP baseline array of %u unsigned shorts",
			xia_dxp_mca->num_baseline_bins );
	}

	/* Allocate memory for the parameter array. */

	xia_status = dxp_max_symbols( &(xia_dxp_mca->system_channel),
					&(xia_dxp_mca->num_parameters) );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to get the number of parameters for MCA '%s'.  "
		"Error code = %d, error status = '%s'",
			mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	xia_dxp_mca->parameter_array = (unsigned short *)
		malloc( xia_dxp_mca->num_parameters * sizeof(unsigned short) );

	if ( xia_dxp_mca->parameter_array == (unsigned short *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocated a DXP parameter array of %hu unsigned shorts",
			xia_dxp_mca->num_parameters );
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_XIA_XERXES */

MX_EXPORT mx_status_type
mxd_xia_dxp_open( MX_RECORD *record )
{
	const char fname[] = "mxd_xia_dxp_open()";

	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_RECORD *xia_dxp_record;
	unsigned long i;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	mca = (MX_MCA *) (record->record_class_struct);

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_dxp_record = xia_dxp_mca->xia_dxp_record;

	MX_DEBUG(-3,("%s invoked for interface '%s' type %lu",
		fname, record->name, xia_dxp_record->mx_type));

	switch( xia_dxp_record->mx_type ) {
	case MXI_GEN_XIA_MDS:
		mx_status = mxd_xia_dxp_mds_open( mca,
					xia_dxp_mca, xia_dxp_record );
		break;

	case MXI_GEN_XIA_NETWORK:
		mx_status = mxd_xia_dxp_network_open( mca,
					xia_dxp_mca, xia_dxp_record );
		break;

	case MXI_GEN_XIA_XERXES:

#if HAVE_XIA_XERXES
		mx_status = mxd_xia_dxp_xerxes_open( mca,
					xia_dxp_mca, xia_dxp_record );
#else
		return mx_error( MXE_UNSUPPORTED, fname,
		"XIA XerXes support is not compiled into this copy of MX.  "
		"You will need to recompile MX including XerXes support "
		"to fix this." );
#endif
		break;

	default:
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"The XIA DXP record '%s' for MCA '%s' is not a Xerxes or "
		"an MDS interface record.", xia_dxp_record->name, record->name);
	}

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Initialize the range of bin numbers used by the MCA
	 * unless we are connected via a 'xia_network' record.
	 */

	if ( xia_dxp_record->mx_type != MXI_GEN_XIA_NETWORK ) {

		if ( xia_dxp_mca->write_parameter == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->write_parameter)( mca,
						"MCALIMLO", 0,
						MXD_XIA_DXP_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_status = (xia_dxp_mca->write_parameter)( mca,
						"MCALIMHI",
						mca->maximum_num_channels - 1L,
						MXD_XIA_DXP_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mca->current_num_channels = mca->maximum_num_channels;
	mca->current_num_rois     = mca->maximum_num_rois;

	for ( i = 0; i < mca->maximum_num_rois; i++ ) {
		mca->roi_array[i][0] = 0;
		mca->roi_array[i][1] = 0;
	}

	/* Initialize the saved preset values to illegal values. */

	xia_dxp_mca->old_preset_type       = (mx_uint32_type) ULONG_MAX;
	xia_dxp_mca->old_preset_high_order = (mx_uint32_type) ULONG_MAX;
	xia_dxp_mca->old_preset_low_order  = (mx_uint32_type) ULONG_MAX;

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_close( MX_RECORD *record )
{
	const char fname[] = "mxd_xia_dxp_close()";

	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_RECORD *xia_dxp_record;
	MX_XIA_MDS *xia_mds;
	MX_XIA_NETWORK *xia_network;
	MX_RECORD **mca_record_array;
	unsigned long i, num_mcas;
	mx_status_type mx_status;

#if HAVE_XIA_XERXES
	MX_XIA_XERXES *xia_xerxes;
#endif

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	mca = (MX_MCA *) (record->record_class_struct);

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Delete our entry in mca_record_array. */

	xia_dxp_record = xia_dxp_mca->xia_dxp_record;

	switch( xia_dxp_record->mx_type ) {
	case MXI_GEN_XIA_MDS:
		xia_mds = (MX_XIA_MDS *) xia_dxp_record->record_type_struct;

		num_mcas = xia_mds->num_mcas;
		mca_record_array = xia_mds->mca_record_array;
		break;

	case MXI_GEN_XIA_NETWORK:
		xia_network = (MX_XIA_NETWORK *)
					xia_dxp_record->record_type_struct;

		num_mcas = xia_network->num_mcas;
		mca_record_array = xia_network->mca_record_array;
		break;

	case MXI_GEN_XIA_XERXES:

#if HAVE_XIA_XERXES
		xia_xerxes = (MX_XIA_XERXES *)
					xia_dxp_record->record_type_struct;

		num_mcas = xia_xerxes->num_mcas;
		mca_record_array = xia_xerxes->mca_record_array;

		if ( xia_dxp_mca->parameter_array != NULL ) {
			mx_free( xia_dxp_mca->parameter_array );
		}
		if ( xia_dxp_mca->baseline_array != NULL ) {
			mx_free( xia_dxp_mca->baseline_array );
		}
		if ( xia_dxp_mca->spectrum_array != NULL ) {
			mx_free( xia_dxp_mca->spectrum_array );
		}
#else
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"You are trying to close MCA '%s' which uses XIA interface '%s' which is "
"of type MXI_GEN_XIA_XERXES.  However, XIA XerXes support is not compiled "
"into this copy of MX.  This should never happen and is definitely a program "
"bug that should be reported.", record->name, xia_dxp_record->name );
#endif
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"You are trying to close MCA '%s' which uses interface record '%s'.  "
"'%s' is not a XIA interface record.",
			mca->record->name,
			xia_dxp_record->name,
			xia_dxp_record->name );
		break;
	}

	for ( i = 0; i < num_mcas; i++ ) {

		if ( record == mca_record_array[i] ) {
			mca_record_array[i] = NULL;

			break;		/* Exit the for() loop. */
		}
	}

	/* Don't treat it as an error if we did not find an entry
	 * in mca_record_array.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_start( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_start()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	double preset_time;
	double xia_clock_ticks, xia_high_order_double, xia_low_order_double;
	mx_uint32_type xia_preset_type;
	mx_uint32_type xia_high_order, xia_low_order;
	mx_status_type mx_status;

#if MXD_XIA_VERIFY_PRESETS
	mx_uint32_type returned_preset_type;
	mx_uint32_type returned_xia_high_order, returned_xia_low_order;
#endif /* MXD_XIA_VERIFY_PRESETS */

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Select the preset type. */

	switch( mca->preset_type ) {
	case MXF_MCA_PRESET_NONE:
		xia_preset_type = MXF_XIA_PRESET_NONE;
		preset_time = 0.0;
		break;

	case MXF_MCA_PRESET_LIVE_TIME:
		xia_preset_type = MXF_XIA_PRESET_LIVE_TIME;
		preset_time = mca->preset_live_time;
		break;

	case MXF_MCA_PRESET_REAL_TIME:
		xia_preset_type = MXF_XIA_PRESET_REAL_TIME;
		preset_time = mca->preset_real_time;
		break;
	default:
		return mx_error( MXE_UNSUPPORTED, fname,
			"Preset type %d is not supported for record '%s'",
			mca->preset_type, mca->record->name );
		break;
	}

#if MXD_XIA_DXP_DEBUG
	MX_DEBUG(-2,("%s: xia_preset_type = %lu", fname,
					(unsigned long) xia_preset_type));
#endif

	/* Compute the time preset in XIA clock ticks of 400 nanoseconds. */

	xia_clock_ticks = preset_time / MX_XIA_DXP_CLOCK_TICK;

	/* The preset time is sent to the MCA controller as a 16-bit
	 * high order quantity plus a 16-bit low order quantity.
	 */

	/* 2^16 = 65536 */

	xia_high_order_double = xia_clock_ticks / 65536.0;

	xia_high_order = (unsigned long) xia_high_order_double;

	xia_low_order_double = xia_clock_ticks
				- 65536.0 * (double) xia_high_order;

	xia_low_order = (unsigned long) xia_low_order_double;

	if ( xia_high_order >= 65536 ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"The requested preset time of %g seconds would exceed the maximum "
	"allowed counting time for the XIA MCA of %g seconds.", preset_time,
			MX_XIA_DXP_CLOCK_TICK * pow( 2.0, 32.0 ) );
	}

	MX_DEBUG(-2,("%s: preset_time = %g", fname, preset_time));
	MX_DEBUG(-2,("%s: xia_high_order = %lu, xia_low_order = %lu",
			fname, xia_high_order, xia_low_order));

	/* If all the preset values for this MCA start are the same as 
	 * the values from the last MCA start, then we do not need to
	 * send any change parameter requests and can go ahead and start
	 * the MCA.
	 */

	if ( ( xia_preset_type == xia_dxp_mca->old_preset_type )
	  && ( xia_high_order  == xia_dxp_mca->old_preset_high_order )
	  && ( xia_low_order   == xia_dxp_mca->old_preset_low_order ) )
	{
		MX_DEBUG(-3,("%s: Presets will NOT be changed.", fname));

		if ( xia_dxp_mca->start_run == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->start_run)( mca, 1,
						MXD_XIA_DXP_DEBUG );

		/* We are done, so return to the caller. */

		return mx_status;
	}

	/* One or more of the preset parameters are different than
	 * before, so we must set them all.
	 */

	MX_DEBUG(-3,("%s: Presets will be changed.", fname));

	/******* Set the preset type. *******/

	if ( xia_dxp_mca->write_parameter_to_all_channels == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	mx_status = (xia_dxp_mca->write_parameter_to_all_channels)( mca,
						"PRESET", xia_preset_type,
						MXD_XIA_DXP_DEBUG );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if MXD_XIA_VERIFY_PRESETS
	/* Read back the preset type to verify it was set correctly. */

	if ( xia_dxp_mca->read_parameter == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	mx_status = (xia_dxp_mca->read_parameter)( mca,
						"PRESET", &returned_preset_type,
						MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( returned_preset_type != xia_preset_type ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"The attempt to set the XIA DXP preset type to %lu failed.  "
		"When the value was read back, it had been set to %lu.",
			(unsigned long) xia_preset_type,
			(unsigned long) returned_preset_type );
	}
#endif /* MXD_XIA_VERIFY_PRESETS */

	xia_dxp_mca->old_preset_type = xia_preset_type;

	/******* Set the timer presets. *******/

	mx_status = (xia_dxp_mca->write_parameter_to_all_channels)( mca,
						"PRESETLEN1", xia_low_order,
						MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = (xia_dxp_mca->write_parameter_to_all_channels)( mca,
						"PRESETLEN0", xia_high_order,
						MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if MXD_XIA_VERIFY_PRESETS
	/* Read back and verify the PRESETLENx values. */

	mx_status = (xia_dxp_mca->read_parameter)( mca,
						"PRESETLEN1",
						&returned_xia_low_order,
						MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = (xia_dxp_mca->read_parameter)( mca,
						"PRESETLEN0",
						&returned_xia_high_order,
						MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s: returned_xia_low_order = %#lx, returned_xia_high_order = %#lx",
		fname, (unsigned long) returned_xia_low_order,
		(unsigned long) returned_xia_high_order));

	if ( (returned_xia_low_order != xia_low_order)
	  || (returned_xia_high_order != xia_high_order) )
	{
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"The attempt to set PRESETLEN0 and PRESETLEN1 to %#lx and %#lx failed.  "
"When the values were read back, they had been set to %#lx and %#lx.",
			(unsigned long) xia_high_order,
			(unsigned long) xia_low_order,
			(unsigned long) returned_xia_high_order,
			(unsigned long) returned_xia_low_order );
	}
#endif /* MXD_XIA_VERIFY_PRESETS */

	xia_dxp_mca->old_preset_high_order = xia_high_order;
	xia_dxp_mca->old_preset_low_order = xia_low_order;

	xia_dxp_mca->new_statistics_available = TRUE;

	/* Start the MCA. */

	if ( xia_dxp_mca->start_run == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	mx_status = (xia_dxp_mca->start_run)( mca, 1, MXD_XIA_DXP_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_stop( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_stop()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Stop the MCA. */

	if ( xia_dxp_mca->stop_run == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	mx_status = (xia_dxp_mca->stop_run)( mca, MXD_XIA_DXP_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_read( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_read()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( xia_dxp_mca->read_spectrum == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	mx_status = (xia_dxp_mca->read_spectrum)( mca, MXD_XIA_DXP_DEBUG );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_clear( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_clear()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* There doesn't appear to be a way of doing this without
	 * starting a run.
	 */

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_busy( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_busy()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( xia_dxp_mca->is_busy == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	mx_status = (xia_dxp_mca->is_busy)( mca,
				&(mca->busy),
				MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( mca->busy == FALSE ) {

		/* Send a stop run command so that the run enabled bit
		 * in the DXP is cleared.
		 */

		if ( xia_dxp_mca->stop_run == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->stop_run)( mca, MXD_XIA_DXP_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	MX_DEBUG( 2,("%s: mca->busy = %d", fname, mca->busy));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_get_parameter( MX_MCA *mca )
{
	mx_status_type mx_status;

	if ( mca->record->mx_type == MXI_GEN_XIA_NETWORK ) {
		mx_status = mxi_xia_network_get_parameter( mca );
	} else {
		mx_status = mxd_xia_dxp_get_local_parameter( mca );
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_set_parameter( MX_MCA *mca )
{
	mx_status_type mx_status;

	if ( mca->record->mx_type == MXI_GEN_XIA_NETWORK ) {
		mx_status = mxi_xia_network_set_parameter( mca );
	} else {
		mx_status = mxd_xia_dxp_set_local_parameter( mca );
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_get_local_parameter( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_get_local_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_uint32_type low_limit, high_limit;
	mx_uint32_type i, j, roi_boundary;
	unsigned long roi[2];
	char name[20];
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for MCA '%s', parameter type '%s' (%d).",
		fname, mca->record->name,
		mx_get_field_label_string(mca->record,mca->parameter_type),
		mca->parameter_type));

	switch( mca->parameter_type ) {
	case MXLV_MCA_CURRENT_NUM_CHANNELS:

		if ( xia_dxp_mca->read_parameter == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->read_parameter)( mca,
						"MCALIMLO", &low_limit,
						MXD_XIA_DXP_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_status = (xia_dxp_mca->read_parameter)( mca,
						"MCALIMHI", &high_limit,
						MXD_XIA_DXP_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mca->current_num_channels = (long) high_limit
						- (long) low_limit + 1L;

		if ( mca->current_num_channels > mca->maximum_num_channels ) {
			mx_warning(
"MCA '%s' reports that it currently has %ld channels.  "
"The MX record is currently configured for a maximum of %ld channels, "
"so channels %ld to %ld will be discarded.  "
"This probably means that the MX database is not correctly configured.",
				mca->record->name,
				mca->current_num_channels,
				mca->maximum_num_channels,
				mca->current_num_channels,
				mca->maximum_num_channels - 1 );

			mca->current_num_channels = mca->maximum_num_channels;
		}
		break;
	case MXLV_MCA_ROI:
		i = mca->roi_number;

		sprintf( name, "SCA%luLO", (unsigned long) i );

		if ( xia_dxp_mca->read_parameter == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->read_parameter)( mca,
						name, &roi_boundary,
						MXD_XIA_DXP_DEBUG );
		mca->roi[0] = roi_boundary;

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( name, "SCA%luHI", (unsigned long) i );

		mx_status = (xia_dxp_mca->read_parameter)( mca,
						name, &roi_boundary,
						MXD_XIA_DXP_DEBUG );

		mca->roi[1] = roi_boundary;

		break;
	case MXLV_MCA_ROI_INTEGRAL:
		/* There seems to be no obvious way to ask for the ROI
		 * integral, so we must fetch the data array and then
		 * compute it ourself.  In that case, why even have
		 * hardware support for ROIs?
		 */

		i = mca->roi_number;

		mx_status = mx_mca_read( mca->record, NULL, NULL );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Make sure the ROI boundaries are correct. */

		mx_status = mx_mca_get_roi( mca->record, i, roi );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG(-2,
	("%s: Integrating MCA '%s' ROI %lu integral from bin %lu to bin %lu",
			fname, mca->record->name, (unsigned long) i,
			roi[0], roi[1]));

		mca->roi_integral = 0;

		for ( j = roi[0]; j <= roi[1]; j++ ) {
			mca->roi_integral += mca->channel_array[j];
		}

		MX_DEBUG(-2,("%s: ROI %lu integral = %lu",
			fname, (unsigned long) i, mca->roi_integral));

		break;
	case MXLV_MCA_LIVE_TIME:
	case MXLV_MCA_REAL_TIME:
		mx_status = mxd_xia_dxp_read_statistics( mca, xia_dxp_mca,
							MXD_XIA_DXP_DEBUG );

		MX_DEBUG(-2,
		("%s: new_statistics_available = %d, mca->busy = %d",
			fname, xia_dxp_mca->new_statistics_available,
			mca->busy ));

		break;
	default:
		return mx_mca_default_get_parameter_handler( mca );
		break;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_set_local_parameter( MX_MCA *mca )
{
	const char fname[] = "mxd_xia_dxp_set_local_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_uint32_type i, low_limit, high_limit;
	char name[20];
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s invoked for MCA '%s', parameter type '%s' (%d).",
		fname, mca->record->name,
		mx_get_field_label_string(mca->record,mca->parameter_type),
		mca->parameter_type));

	if ( xia_dxp_mca->write_parameter == NULL ) {
		return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
			xia_dxp_mca->xia_dxp_record->name );
	}

	switch( mca->parameter_type ) {
	case MXLV_MCA_CURRENT_NUM_CHANNELS:

		if ( xia_dxp_mca->read_parameter == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->read_parameter)( mca,
						"MCALIMLO", &low_limit,
						MXD_XIA_DXP_DEBUG );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		high_limit = (unsigned long)low_limit
				+ mca->current_num_channels - 1L;

		mx_status = (xia_dxp_mca->write_parameter)( mca,
						"MCALIMHI", high_limit,
						MXD_XIA_DXP_DEBUG );
	case MXLV_MCA_ROI:
		i = mca->roi_number;

		sprintf( name, "SCA%luLO", (unsigned long) i );

		mx_status = (xia_dxp_mca->write_parameter)( mca,
						name, mca->roi[0],
						MXD_XIA_DXP_DEBUG );
		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( name, "SCA%luHI", (unsigned long) i );

		mx_status = (xia_dxp_mca->write_parameter)( mca,
						name, mca->roi[1],
						MXD_XIA_DXP_DEBUG );
		break;

	case MXLV_MCA_ROI_ARRAY:
		for ( i = 0; i < mca->current_num_rois; i++ ) {

			sprintf( name, "SCA%luLO", (unsigned long) i );

			mx_status = (xia_dxp_mca->write_parameter)( mca,
						name, mca->roi_array[i][0],
						MXD_XIA_DXP_DEBUG );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			sprintf( name, "SCA%luHI", (unsigned long) i );

			mx_status = (xia_dxp_mca->write_parameter)( mca,
						name, mca->roi_array[i][1],
						MXD_XIA_DXP_DEBUG );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;
		}
		break;
	default:
		return mx_mca_default_set_parameter_handler( mca );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_special_processing_setup( MX_RECORD *record )
{
	const char fname[] = "mxd_xia_dxp_special_processing_setup()";

	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG(2,("%s invoked.", fname));

	record_field_array = record->record_field_array;

	for ( i = 0; i < record->num_record_fields; i++ ) {

		record_field = &record_field_array[i];

		switch( record_field->label_value ) {
		case MXLV_XIA_DXP_STATISTICS:
		case MXLV_XIA_DXP_PARAMETER_VALUE:
		case MXLV_XIA_DXP_PARAM_VALUE_TO_ALL_CHANNELS:
			record_field->process_function
					    = mxd_xia_dxp_process_function;
			break;
		default:
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_xia_dxp_get_mca_array( MX_RECORD *xia_dxp_record,
				unsigned long *num_mcas,
				MX_RECORD ***mca_record_array )
{
	const char fname[] = "mxd_xia_dxp_get_mca_array()";

	MX_XIA_MDS *xia_mds;
	MX_XIA_NETWORK *xia_network;

#if HAVE_XIA_XERXES
	MX_XIA_XERXES *xia_xerxes;
#endif

	if ( xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The xia_dxp_record pointer passed was NULL." );
	}
	if ( num_mcas == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The num_mcas pointer passed was NULL." );
	}
	if ( mca_record_array == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The mca_record_array pointer passed was NULL." );
	}

	switch( xia_dxp_record->mx_type ) {

	case MXI_GEN_XIA_MDS:
		xia_mds = (MX_XIA_MDS *) xia_dxp_record->record_type_struct;

		if ( xia_mds == (MX_XIA_MDS *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_XIA_MDS pointer for record '%s' is NULL.",
				xia_dxp_record->name );
		}

		*num_mcas = xia_mds->num_mcas;
		*mca_record_array = xia_mds->mca_record_array;
		break;

	case MXI_GEN_XIA_NETWORK:
		xia_network = (MX_XIA_NETWORK *)
					xia_dxp_record->record_type_struct;

		if ( xia_network == (MX_XIA_NETWORK *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_XIA_NETWORK pointer for record '%s' is NULL.",
				xia_dxp_record->name );
		}

		*num_mcas = xia_network->num_mcas;
		*mca_record_array = xia_network->mca_record_array;
		break;

#if HAVE_XIA_XERXES
	case MXI_GEN_XIA_XERXES:
		xia_xerxes = (MX_XIA_XERXES *)
					xia_dxp_record->record_type_struct;

		if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_XIA_XERXES pointer for record '%s' is NULL.",
				xia_dxp_record->name );
		}

		*num_mcas = xia_xerxes->num_mcas;
		*mca_record_array = xia_xerxes->mca_record_array;
		break;
#endif

	default:
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"xia_dxp_record '%s' is not a "
		"XIA MDS record or a XIA Xerxes record.",
			xia_dxp_record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_get_corrected_roi( MX_MCA *mca,
			unsigned long roi_number,
			double *corrected_roi_value )
{
	const char fname[] = "mxd_xia_dxp_get_corrected_roi()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	unsigned long mca_value;
	double corrected_value, multiplier;
	mx_status_type mx_status;

	mx_status = mxd_xia_dxp_get_pointers( mca, &xia_dxp_mca, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( corrected_roi_value == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The corrected_roi_error pointer passed was NULL." );
	}

	mx_status = mx_mca_get_roi_integral( mca->record,
						roi_number, &mca_value );
	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	corrected_value = (double) mca_value;

	mx_status = mxd_xia_dxp_read_statistics( mca, xia_dxp_mca,
						MXD_XIA_DXP_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	multiplier = mx_divide_safely( xia_dxp_mca->input_count_rate,
					xia_dxp_mca->output_count_rate );

	corrected_value = mx_multiply_safely( corrected_value, multiplier );

	multiplier = mx_divide_safely( mca->real_time, mca->live_time );

	corrected_value = mx_multiply_safely( corrected_value, multiplier );

	*corrected_roi_value = corrected_value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_xia_dxp_read_statistics( MX_MCA *mca,
				MX_XIA_DXP_MCA *xia_dxp_mca,
				int debug_flag )
{
	const char fname[] = "mxd_xia_dxp_read_statistics()";

	int read_statistics;
	mx_status_type mx_status;

	if ( mca->mca_flags & MXF_MCA_NO_READ_OPTIMIZATION ) {
		read_statistics = TRUE;

	} else if ( xia_dxp_mca->new_statistics_available ) {
		read_statistics = TRUE;

	} else {
		read_statistics = FALSE;
	}

	MX_DEBUG(-2,("%s: read_statistics = %d", fname, read_statistics));

	if ( read_statistics ) {

		if ( xia_dxp_mca->read_statistics == NULL ) {
			return mx_error( MXE_INITIALIZATION_ERROR, fname,
		"XIA interface record '%s' has not been initialized properly.",
				xia_dxp_mca->xia_dxp_record->name );
		}

		mx_status = (xia_dxp_mca->read_statistics)( mca, debug_flag );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( mca->busy == FALSE ) {
			xia_dxp_mca->new_statistics_available = FALSE;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

#ifndef MX_PROCESS_GET

#define MX_PROCESS_GET	1
#define MX_PROCESS_PUT	2

#endif


static mx_status_type
mxd_xia_dxp_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxd_xia_dxp_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	mx_uint32_type parameter_value;
	mx_status_type mx_status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;
	mca = (MX_MCA *) record->record_class_struct;
	xia_dxp_mca = (MX_XIA_DXP_MCA *) record->record_type_struct;

	MX_DEBUG(-2,("**** %s invoked, operation = %d, label_value = %ld ****",
		fname, operation, record_field->label_value));

	mx_status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( record_field->label_value ) {
		case MXLV_XIA_DXP_PARAMETER_VALUE:
			if ( xia_dxp_mca->read_parameter == NULL ) {
				return mx_error( MXE_INITIALIZATION_ERROR,
					fname,
		"XIA interface record '%s' has not been initialized properly.",
					xia_dxp_mca->xia_dxp_record->name );
			}

			mx_status = (xia_dxp_mca->read_parameter)( mca,
						xia_dxp_mca->parameter_name,
						&parameter_value,
						MXD_XIA_DXP_DEBUG );

			xia_dxp_mca->parameter_value = parameter_value;

			MX_DEBUG(-2,
		("%s: Read mca '%s' parameter name = '%s', value = %lu",
				fname, mca->record->name,
				xia_dxp_mca->parameter_name,
				xia_dxp_mca->parameter_value));

			break;
		case MXLV_XIA_DXP_STATISTICS:

			mx_status = mxd_xia_dxp_read_statistics(
							mca, xia_dxp_mca,
							MXD_XIA_DXP_DEBUG );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			xia_dxp_mca->statistics[ MX_XIA_DXP_REAL_TIME ]
						= mca->real_time;

			xia_dxp_mca->statistics[ MX_XIA_DXP_LIVE_TIME ]
						= mca->live_time;

			xia_dxp_mca->statistics[ MX_XIA_DXP_INPUT_COUNT_RATE ]
						= xia_dxp_mca->input_count_rate;

			xia_dxp_mca->statistics[ MX_XIA_DXP_OUTPUT_COUNT_RATE ]
					= xia_dxp_mca->output_count_rate;

			xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_FAST_PEAKS ]
						= xia_dxp_mca->num_fast_peaks;

			xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_EVENTS ]
						= xia_dxp_mca->num_events;

			xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_UNDERFLOWS ]
						= xia_dxp_mca->num_underflows;

			xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_OVERFLOWS ]
						= xia_dxp_mca->num_overflows;
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, record_field->label_value));
			break;
		}
		break;
	case MX_PROCESS_PUT:
		switch( record_field->label_value ) {
		case MXLV_XIA_DXP_PARAMETER_VALUE:

			MX_DEBUG(-2,
		("%s: Write mca '%s' parameter name = '%s', value = %lu",
				fname, mca->record->name,
				xia_dxp_mca->parameter_name,
				xia_dxp_mca->parameter_value));

			if ( xia_dxp_mca->write_parameter == NULL ) {
				return mx_error( MXE_INITIALIZATION_ERROR,
					fname,
		"XIA interface record '%s' has not been initialized properly.",
					xia_dxp_mca->xia_dxp_record->name );
			}

			mx_status = (xia_dxp_mca->write_parameter)( mca,
				xia_dxp_mca->parameter_name,
				(mx_uint32_type) xia_dxp_mca->parameter_value,
				MXD_XIA_DXP_DEBUG );

			break;
		case MXLV_XIA_DXP_PARAM_VALUE_TO_ALL_CHANNELS:
			MX_DEBUG(-2,
("%s: Write to all channels for mca '%s' parameter name = '%s', value = %lu",
				fname, mca->record->name,
				xia_dxp_mca->parameter_name,
				xia_dxp_mca->param_value_to_all_channels));

			if ( xia_dxp_mca->write_parameter_to_all_channels
				== NULL )
			{
				return mx_error( MXE_INITIALIZATION_ERROR,
					fname,
		"XIA interface record '%s' has not been initialized properly.",
					xia_dxp_mca->xia_dxp_record->name );
			}

			mx_status =
			    (xia_dxp_mca->write_parameter_to_all_channels)(
				mca,
				xia_dxp_mca->parameter_name,
		(mx_uint32_type) xia_dxp_mca->param_value_to_all_channels,
				MXD_XIA_DXP_DEBUG );

			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, record_field->label_value));
			break;
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %d", operation );
		break;
	}

	return mx_status;
}

#endif /* USE_TCPIP || HAVE_XIA_XERXES */

