/*
 * Name:    i_xia_mds.c
 *
 * Purpose: MX driver for the X-Ray Instrumentation Associates Mesa2x
 *          Data Server used by the DXP-2X multichannel analyzer.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_osdef.h"

#if USE_TCPIP

#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_types.h"
#include "mx_socket.h"
#include "mx_mca.h"
#include "i_xia_mds.h"
#include "d_xia_dxp_mca.h"

MX_RECORD_FUNCTION_LIST mxi_xia_mds_record_function_list = {
	mxi_xia_mds_initialize_type,
	mxi_xia_mds_create_record_structures,
	mxi_xia_mds_finish_record_initialization,
	mxi_xia_mds_delete_record,
	NULL,
	mxi_xia_mds_read_parms_from_hardware,
	mxi_xia_mds_write_parms_to_hardware,
	mxi_xia_mds_open,
	mxi_xia_mds_close
};

MX_RECORD_FIELD_DEFAULTS mxi_xia_mds_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_XIA_MDS_STANDARD_FIELDS
};

long mxi_xia_mds_num_record_fields
		= sizeof( mxi_xia_mds_record_field_defaults )
			/ sizeof( mxi_xia_mds_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_xia_mds_rfield_def_ptr
			= &mxi_xia_mds_record_field_defaults[0];

#define MXI_XIA_MDS_DEBUG			TRUE

#define MXI_XIA_MDS_SHOW_NETWORK_BYTE_ORDER	FALSE

/* A private function for the use of the driver. */

static mx_status_type
mxi_xia_mds_get_pointers( MX_MCA *mca,
			MX_XIA_DXP_MCA **xia_dxp_mca,
			MX_XIA_MDS **xia_mds,
			const char *calling_fname )
{
	const char fname[] = "mxi_xia_mds_get_pointers()";

	MX_RECORD *xia_dxp_record;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_dxp_mca == (MX_XIA_DXP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_DXP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_mds == (MX_XIA_MDS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_MDS pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*xia_dxp_mca = (MX_XIA_DXP_MCA *) mca->record->record_type_struct;

	xia_dxp_record = (*xia_dxp_mca)->xia_dxp_record;

	if ( xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The xia_dxp_record pointer for MCA record '%s' is NULL.",
			mca->record->name );
	}

	*xia_mds = (MX_XIA_MDS *) xia_dxp_record->record_type_struct;

	if ( (*xia_mds) == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_MDS pointer for XIA DXP record '%s' is NULL.",
			xia_dxp_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_xia_mds_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_mds_create_record_structures()";

	MX_XIA_MDS *xia_mds;

	/* Allocate memory for the necessary structures. */

	xia_mds = (MX_XIA_MDS *) malloc( sizeof(MX_XIA_MDS) );

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_XIA_MDS structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;;
	record->record_type_struct = xia_mds;

	record->class_specific_function_list = NULL;

	xia_mds->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_mds_finish_record_initialization()";

	MX_XIA_MDS *xia_mds;

	xia_mds = (MX_XIA_MDS *) record->record_type_struct;

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_MDS pointer for record '%s' is NULL.",
			record->name );
	}

	xia_mds->socket = MX_INVALID_SOCKET;

	/* Allocate memory for the send buffer. */

	xia_mds->send_buffer = (mx_uint32_type *) malloc(
		MX_XIA_MDS_MAXIMUM_BUFFER_LENGTH * sizeof(mx_uint32_type) );

	if ( xia_mds->send_buffer == (mx_uint32_type *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %u element array "
		"of 32-bit integers for the XIA MDS send buffer.",
			MX_XIA_MDS_MAXIMUM_BUFFER_LENGTH );
	}

	/* Allocate memory for the receive buffer. */

	xia_mds->receive_buffer = (mx_uint32_type *) malloc(
		MX_XIA_MDS_MAXIMUM_BUFFER_LENGTH * sizeof(mx_uint32_type) );

	if ( xia_mds->receive_buffer == (mx_uint32_type *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %u element array "
		"of 32-bit integers for the XIA MDS receive buffer.",
			MX_XIA_MDS_MAXIMUM_BUFFER_LENGTH );
	}

	/* Initialize the request id. */

	xia_mds->request_id = 100;

	/* We cannot set up the restore_configuration_response and
	 * mca_record_array structures until mxi_xia_mds_open() is
	 * invoked, so for now we initialize these data structures
	 * to all zeros.
	 */

	xia_mds->num_mcas = 0;
	xia_mds->mca_record_array = NULL;
	xia_mds->restore_configuration_response = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_delete_record( MX_RECORD *record )
{
	MX_XIA_MDS *xia_mds;

        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }

	xia_mds = (MX_XIA_MDS *) record->record_type_struct;

        if ( xia_mds != NULL ) {

		if ( xia_mds->send_buffer != NULL ) {
			free( xia_mds->send_buffer );
		}
		if ( xia_mds->receive_buffer != NULL ) {
			free( xia_mds->receive_buffer );
		}
		if ( xia_mds->mca_record_array != NULL ) {
			free( xia_mds->mca_record_array );
		}
		if ( xia_mds->restore_configuration_response != NULL ) {
			free( xia_mds->restore_configuration_response );
		}

                free( xia_mds );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_open( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_mds_open()";

	MX_XIA_MDS *xia_mds;
	unsigned long i, num_mcas;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_mds = (MX_XIA_MDS *) (record->record_type_struct);

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_MDS pointer for record '%s' is NULL.", record->name);
	}

	/* Open the connection to the XIA Mesa2x Data Server. */

	mx_status = mx_socket_open_as_client( &(xia_mds->socket), 
			xia_mds->hostname, xia_mds->port_number, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Check to see if the connection is working by having the server
	 * echo back some characters that we send it.
	 */

	mx_status = mxi_xia_mds_echo_test( record, "abcdefghijklmnop",
						MXI_XIA_MDS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Restore the detector configuration. */

	mx_status = mxi_xia_mds_restore_configuration( record,
					xia_mds->config_filename,
					MXI_XIA_MDS_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The 'Restore Configuration From File' command returns the
	 * number of installed MCAs plus the location of all the
	 * installed MCAs.  This gives us enough information to 
	 * allocate arrays for the individual MCA records.
	 */

	num_mcas = xia_mds->receive_buffer[ MXH_XIA_MESSAGE_BODY ];

	xia_mds->mca_record_array = ( MX_RECORD ** ) malloc(
					num_mcas * sizeof( MX_RECORD * ) );

	if ( xia_mds->mca_record_array == ( MX_RECORD ** ) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %lu element array "
		"of MX_RECORD pointers for the mca_record_array data "
		"structure of record '%s'.", num_mcas, record->name );
	}

	for ( i = 0; i < num_mcas; i++ ) {
		xia_mds->mca_record_array[i] = NULL;
	}

	/* Save a copy of the XIA configuration for use by the open()
	 * routines of the individual MCAs.
	 */

	xia_mds->restore_configuration_response = ( unsigned long * ) malloc(
				3 * num_mcas * sizeof( unsigned long ) );

	if ( xia_mds->restore_configuration_response == NULL ) {

		free( xia_mds->mca_record_array );

		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %lu element array "
		"of unsigned longs to contain the 'Restore Configuration "
		"From File' response for record '%s'.",
			3 * num_mcas, record->name);
	}

	for ( i = 0; i < 3 * num_mcas; i++ ) {
		xia_mds->restore_configuration_response[i] =
			xia_mds->receive_buffer[ i + MXH_XIA_MESSAGE_BODY + 1 ];
	}

	xia_mds->num_mcas = num_mcas;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_close( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_mds_close()";

	MX_XIA_MDS *xia_mds;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_mds = (MX_XIA_MDS *) (record->record_type_struct);

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_MDS pointer for record '%s' is NULL.", record->name);
	}

	/* Close the connection to the XIA Mesa2x Data Server. */

	if ( mx_socket_is_open( xia_mds->socket ) ) {
		mx_status = mx_socket_close( xia_mds->socket );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxi_xia_mds_echo_test( MX_RECORD *record, char *test_data, int debug_flag )
{
	const char fname[] = "mxi_xia_mds_echo_test()";

	MX_XIA_MDS *xia_mds;
	int num_test_chars;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	xia_mds = (MX_XIA_MDS *) record->record_type_struct;

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_MDS pointer for record '%s' is NULL.",
			record->name );
	}

	num_test_chars = strlen( test_data );

	if ( num_test_chars % 4 != 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The number of characters %d in test data '%s' is not a multiple of 4.",
			num_test_chars, test_data );
	}

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ] = 12 + num_test_chars;
	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_ECHO_TEST;

	memcpy( (void *) &(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY]),
			test_data, num_test_chars );

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( memcmp( test_data,
		(void *) &(xia_mds->receive_buffer[ MXH_XIA_MESSAGE_BODY ]),
		num_test_chars ) != 0 )
	{
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"The characters '%s' echoed by the server do not match the "
		"characters '%s' that were sent.",
		  (char *) &(xia_mds->receive_buffer[ MXH_XIA_MESSAGE_BODY ]),
		  test_data );
	}

	return MX_SUCCESSFUL_RESULT;;
}

MX_EXPORT mx_status_type
mxi_xia_mds_restore_configuration( MX_RECORD *record,
					char *config_filename,
					int debug_flag )
{
	const char fname[] = "mxi_xia_mds_restore_configuration()";

	MX_XIA_MDS *xia_mds;
	unsigned long num_words_packed, maximum_word_length;
	mx_uint32_type error_code;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	xia_mds = (MX_XIA_MDS *) record->record_type_struct;

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_MDS pointer for record '%s' is NULL.",
			record->name );
	}

	MX_DEBUG( 2,("%s invoked for config file '%s'.",
			fname, config_filename));

	/* Construct the "Restore Configuration From File" command. */

	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ]
					= MXF_XIA_RESTORE_CONFIGURATION;

	maximum_word_length = (MX_XIA_MDS_MAXIMUM_BODY_LENGTH / 4) - 1;

	mx_status = mxi_xia_mds_pack_string(
			&(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 1 ] ),
		 	&num_words_packed, maximum_word_length,
			config_filename );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY ]
			= (unsigned long) strlen(config_filename);

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ]
		= MX_XIA_MDS_HEADER_LENGTH + 4 * (num_words_packed + 1);

	if ( debug_flag ) {
		mx_status = mxi_xia_mds_display_packed_string(
			&(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 1 ]),
			num_words_packed );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* Send the command. */

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code == 0 ) {
		return MX_SUCCESSFUL_RESULT;
	} else
	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error loading configuration file.",
			error_code );
	} else
	if ( error_code & 0x2 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error reading/downloading DSP code.",
			error_code );
	} else
	if ( error_code & 0x4 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error reading/downloading Fippi code.",
			error_code );
	} else
	if ( error_code & 0x8 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error testing memory.",
			error_code );
	} else
	if ( error_code & 0x10 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error initializing CAMAC controller.",
			error_code );
	} else {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"MDS error code %#lx returned:  Unrecognized error code.",
			error_code );
	}
}

MX_EXPORT mx_status_type
mxi_xia_mds_is_busy( MX_MCA *mca, int *busy_flag, int debug_flag )
{
	const char fname[] = "mxi_xia_mds_is_busy()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_uint32_type general_status_register, error_code;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* Construct the 'Read Status' command. */

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ]
					= MX_XIA_MDS_HEADER_LENGTH + 4;

	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_READ_STATUS;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY ]
						= xia_dxp_mca->dxp_module;

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Invalid module number.",
			error_code );
	} else
	if ( error_code != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"MDS error code %#lx returned:  Unrecognized error code.",
			error_code );
	}

	general_status_register
		= xia_mds->receive_buffer[ MXH_XIA_MESSAGE_BODY ];

	/* Mask off all but the low order bits. */

	general_status_register &= 0xffff;

	if ( general_status_register & (1 << (xia_dxp_mca->dxp_channel)) ) {

		*busy_flag = TRUE;
	} else {
		*busy_flag = FALSE;
	}

	MX_DEBUG( 2,("%s: general_status_register = %#lx, *busy_flag = %d",
				fname, general_status_register, *busy_flag));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_read_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type *value_ptr,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_read_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	unsigned long num_words_packed, maximum_word_length;
	mx_uint32_type error_code;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_READ_PARAMETER;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY ] = xia_dxp_mca->crate;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 1 ] = xia_dxp_mca->slot;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 2 ]
						= xia_dxp_mca->dxp_channel;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 3 ]
						= strlen(parameter_name);

	maximum_word_length = (MX_XIA_MDS_MAXIMUM_BODY_LENGTH / 4) - 4;

	mx_status = mxi_xia_mds_pack_string(
			&(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 4 ]),
			&num_words_packed, maximum_word_length,
			parameter_name );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		mx_status = mxi_xia_mds_display_packed_string(
			&(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 4 ]),
			num_words_packed );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ]
			= MX_XIA_MDS_HEADER_LENGTH + 4 * (num_words_packed + 4);

	/* Send the command. */

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"MDS error code %#lx returned:  Error reading from parameter '%s'.",
			error_code, parameter_name );
	} else
	if ( error_code & 0x2 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"MDS error code %#lx returned:  Invalid crate/slot/channel for parameter '%s'.",
			error_code, parameter_name );
	} else if ( error_code != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"MDS error code %#lx returned:  Unrecognized error code for parameter '%s'.",
			error_code, parameter_name );
	}

	*value_ptr = xia_mds->receive_buffer[ MXH_XIA_MESSAGE_BODY ];

	MX_DEBUG( 2,("%s: value = %ld", fname, *value_ptr));

	return MX_SUCCESSFUL_RESULT;
}


static mx_status_type
mxi_xia_mds_write_channel_parameter(
			MX_XIA_DXP_MCA *xia_dxp_mca,
			MX_XIA_MDS *xia_mds,
			mx_uint32_type dxp_channel,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_write_channel_parameter()";

	unsigned long num_words_packed, maximum_word_length;
	mx_uint32_type error_code;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_WRITE_PARAMETER;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY ] = xia_dxp_mca->crate;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 1 ] = xia_dxp_mca->slot;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 2 ] = dxp_channel;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 3 ] = value;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 4 ]
						= strlen(parameter_name);

	maximum_word_length = (MX_XIA_MDS_MAXIMUM_BODY_LENGTH / 4) - 5;

	mx_status = mxi_xia_mds_pack_string(
			&(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 5 ]),
			&num_words_packed, maximum_word_length,
			parameter_name );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		mx_status = mxi_xia_mds_display_packed_string(
			&(xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 5 ]),
			num_words_packed );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ]
			= MX_XIA_MDS_HEADER_LENGTH + 4 * (num_words_packed + 5);

	/* Send the command. */

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"MDS error code %#lx returned:  Error writing value %lu for parameter '%s'.",
			error_code, (unsigned long) value, parameter_name );
	} else
	if ( error_code & 0x2 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"MDS error code %#lx returned:  Invalid crate/slot/channel for parameter '%s'.",
			error_code, parameter_name );
	} else
	if ( error_code & 0x4 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"MDS error code %#lx returned:  Value %lu is out of range for parameter '%s'.",
			error_code, (unsigned long) value, parameter_name );
	} else if ( error_code != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"MDS error code %#lx returned:  Unrecognized error code for parameter '%s'.",
			error_code, parameter_name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_write_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_write_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_mds_write_channel_parameter(
					xia_dxp_mca,
					xia_mds,
					xia_dxp_mca->dxp_channel,
					parameter_name,
					value,
					debug_flag );
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_xia_mds_write_parameter_to_all_channels( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_write_parameter_to_all_channels()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_mds_write_channel_parameter(
					xia_dxp_mca,
					xia_mds,
					MXF_XIA_ALL_CHANNELS,
					parameter_name,
					value,
					debug_flag );
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_xia_mds_start_run( MX_MCA *mca,
			int clear_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_start_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_uint32_type error_code;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ]
					= MX_XIA_MDS_HEADER_LENGTH + 4;

	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_START_RUN;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY ] = clear_flag;

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error starting run.",
			error_code );
	} else if ( error_code != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"MDS error code %#lx returned:  Unrecognized error code.",
			error_code );
	}

	return MX_SUCCESSFUL_RESULT;
}


MX_EXPORT mx_status_type
mxi_xia_mds_stop_run( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_stop_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_uint32_type error_code;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ] = MX_XIA_MDS_HEADER_LENGTH;
	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_STOP_RUN;

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error stopping run.",
			error_code );
	} else if ( error_code != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"MDS error code %#lx returned:  Unrecognized error code.",
			error_code );
	}

	return MX_SUCCESSFUL_RESULT;
}


static mx_status_type
mxi_xia_mds_read_data( MX_MCA *mca,
			int data_type,
			mx_uint32_type **data_ptr,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_read_data()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_uint32_type error_code;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	xia_mds->send_buffer[ MXH_XIA_REQUEST_TYPE ] = MXF_XIA_READ_DATA;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY ] = xia_dxp_mca->crate;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 1 ] = xia_dxp_mca->slot;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 2 ]
						= xia_dxp_mca->dxp_channel;
	xia_mds->send_buffer[ MXH_XIA_MESSAGE_BODY + 3 ] = data_type;

	switch( data_type ) {
	case MXF_XIA_READ_MCA_SPECTRUM:
	case MXF_XIA_READ_STATISTICS:
	case MXF_XIA_READ_BASELINE_SPECTRUM:
		xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ]
					= MX_XIA_MDS_HEADER_LENGTH + 4 * 4;
		break;
	case MXF_XIA_READ_ADC_TRACE:
	case MXF_XIA_READ_EXTERNAL_MEMORY:
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Support for data type %d is not yet implemented.", data_type);
		break;
	}

	mx_status = mxi_xia_mds_command( xia_mds, debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	error_code = xia_mds->receive_buffer[ MXH_XIA_ERROR_CODE ];

	if ( error_code & 0x1 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Invalid data type.",
			error_code );
	} else
	if ( error_code & 0x2 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error reading the MCA spectrum.",
			error_code );
	} else
	if ( error_code & 0x4 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  No bins in the MCA spectrum.",
			error_code );
	} else
	if ( error_code & 0x8 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error reading the baseline spectrum.",
			error_code );
	} else
	if ( error_code & 0x10 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
"MDS error code %#lx returned:  Error setting parameters for ADC trace.",
			error_code );
	} else
	if ( error_code & 0x20 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Timeout waiting for BUSY.",
			error_code );
	} else
	if ( error_code & 0x40 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Error reading the history buffer.",
			error_code );
	} else
	if ( error_code & 0x80 ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
	"MDS error code %#lx returned:  Invalid crate/slot/channel.",
			error_code );
	} else if ( error_code != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"MDS error code %#lx returned:  Unrecognized error code.",
			error_code );
	}

	*data_ptr = &( xia_mds->receive_buffer[ MXH_XIA_MESSAGE_BODY ] );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_read_spectrum( MX_MCA *mca, int debug_flag )
{
	mx_uint32_type *local_ptr, *local_array;
	unsigned long i;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_read_data( mca,
				MXF_XIA_READ_MCA_SPECTRUM,
				&local_ptr,
				debug_flag );

	/* The data in the spectrum starts at the tenth array element. */

	local_array = &(local_ptr[9]);

	/* Copy the data to the supplied data array. */

	for ( i = 0; i < mca->current_num_channels; i++ ) {
		mca->channel_array[i] = (unsigned long) local_array[i];
	}

	return mx_status;
}

#define TWO_TO_THE_32ND_POWER	4294967296.0

MX_EXPORT mx_status_type
mxi_xia_mds_read_statistics( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_mds_read_statistics()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_MDS *xia_mds;
	mx_uint32_type *local_ptr;
	mx_status_type mx_status;

	mx_status = mxi_xia_mds_get_pointers( mca,
					&xia_dxp_mca, &xia_mds, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxi_xia_mds_read_data( mca,
				MXF_XIA_READ_STATISTICS,
				&local_ptr,
				debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mca->real_time = MX_XIA_DXP_CLOCK_TICK * ( (double)local_ptr[2]
			+ TWO_TO_THE_32ND_POWER * (double)local_ptr[3] );

	mca->live_time = MX_XIA_DXP_CLOCK_TICK * ( (double)local_ptr[0]
			+ TWO_TO_THE_32ND_POWER * (double)local_ptr[1] );

	xia_dxp_mca->num_fast_peaks = local_ptr[7];

	xia_dxp_mca->input_count_rate = mx_divide_safely(
					(double) xia_dxp_mca->num_fast_peaks,
						mca->live_time );

	xia_dxp_mca->num_events = local_ptr[4];

	xia_dxp_mca->output_count_rate = mx_divide_safely(
					(double) xia_dxp_mca->num_events,
						mca->live_time );

	MX_DEBUG(-3,("%s: num_events = %lu, num_fast_peaks = %lu", fname,
		(unsigned long) xia_dxp_mca->num_events,
		(unsigned long) xia_dxp_mca->num_fast_peaks));

	xia_dxp_mca->num_underflows = local_ptr[5];
	xia_dxp_mca->num_overflows = local_ptr[6];

	MX_DEBUG(-3,("%s: num_underflows = %lu, num_overflows = %lu", fname,
		(unsigned long) xia_dxp_mca->num_underflows,
		(unsigned long) xia_dxp_mca->num_overflows));

	return mx_status;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxi_xia_mds_command( MX_XIA_MDS *xia_mds, int debug_flag )
{
	const char fname[] = "mxi_xia_mds_command()";

	unsigned long i, temp;
	unsigned long message_bytes, remaining_message_bytes, message_words;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( xia_mds == (MX_XIA_MDS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_XIA_MDS pointer passed was NULL." );
	}

	xia_mds->request_id++;

	xia_mds->send_buffer[ MXH_XIA_REQUEST_ID ] = xia_mds->request_id;

	message_bytes = xia_mds->send_buffer[ MXH_XIA_MESSAGE_SIZE ];

	if ( ( message_bytes % 4 ) != 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The number of bytes (%ld) in the command is not a multiple of 4.",
			message_bytes );
	}

	/* Convert the message to network byte order. */

	message_words = message_bytes / 4;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: *********** Start ***********", fname));
	}

	for ( i = 0; i < message_words; i++ ) {

		temp = htonl( xia_mds->send_buffer[i] );

		if ( debug_flag ) {

#if MX_XIA_MDS_SHOW_NETWORK_BYTE_ORDER
			MX_DEBUG(-2,("%s: send_buffer[%lu]: %lx -> %lx",
				fname, i, xia_mds->send_buffer[i], temp));
#else
			MX_DEBUG(-2,("%s: send_buffer[%lu] = %lx",
				fname, i, xia_mds->send_buffer[i]));
#endif
		}
		xia_mds->send_buffer[i] = temp;
	}

	/* Transmit the command. */

	mx_status = mx_socket_send( xia_mds->socket,
					xia_mds->send_buffer,
					message_bytes, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Find out how many bytes are in the response by reading the
	 * first 32 bit integer in the response.
	 */

	mx_status = mx_socket_receive( xia_mds->socket,
					xia_mds->receive_buffer,
					sizeof(mx_uint32_type), 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	message_bytes = ntohl( xia_mds->receive_buffer[ MXH_XIA_MESSAGE_SIZE ]);

	if ( message_bytes > MX_XIA_MDS_MAXIMUM_BUFFER_LENGTH ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The length %lu bytes of the response from the XIA server "
		"is too long to fit in the buffer of %lu bytes.",
		message_bytes,
		(unsigned long) MX_XIA_MDS_MAXIMUM_BUFFER_LENGTH );
	}

	/* Get the rest of the response. */

	remaining_message_bytes = message_bytes - sizeof(mx_uint32_type);

	mx_status = mx_socket_receive( xia_mds->socket,
					&(xia_mds->receive_buffer[1]),
					remaining_message_bytes, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Convert the message to host byte order. */

	message_words = message_bytes / 4;

	for ( i = 0; i < message_words; i++ ) {

		temp = ntohl( xia_mds->receive_buffer[i] );

		if ( debug_flag ) {
			if ( i < 100 ) {
#if MX_XIA_MDS_SHOW_NETWORK_BYTE_ORDER
			    MX_DEBUG(-2,("%s: receive_buffer[%lu]: %lx -> %lx",
				fname, i, xia_mds->receive_buffer[i], temp));
#else
			    MX_DEBUG(-2,("%s: receive_buffer[%lu] = %lx",
				fname, i, temp));
#endif
			} else if ( i == 100 ) {
			    MX_DEBUG(-2,
			    ("%s: *** message display truncated ***", fname ));
			} else {
				/* Do nothing. */
			}
		}
		xia_mds->receive_buffer[i] = temp;
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: ************ End ************", fname));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_pack_string( mx_uint32_type *destination,
			unsigned long *num_words_packed,
			unsigned long maximum_num_words,
			char *source )
{
	const char fname[] = "mxi_xia_mds_pack_string()";

	unsigned long i, j, k, shift, string_length;
	unsigned long truncated_num_words, packed_num_words, extra_bytes;

	if ( source == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The source string pointer passed was NULL." );
	}

	MX_DEBUG( 2,("%s: source = '%s'", fname, source));

	string_length = strlen( source );

	truncated_num_words = string_length / 4;

	extra_bytes = string_length % 4;

	if ( extra_bytes == 0 ) {
		packed_num_words = truncated_num_words;
	} else {
		packed_num_words = truncated_num_words + 1;
	}

	MX_DEBUG( 2,("%s: string_length = %lu",
			fname, (unsigned long) string_length));
	MX_DEBUG( 2,("%s: truncated_num_words = %lu, packed_num_words = %lu",
			fname, truncated_num_words, packed_num_words));
	MX_DEBUG( 2,("%s: extra_bytes = %lu", fname, extra_bytes));

	if ( packed_num_words > maximum_num_words ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"It would take %lu 32-bit words to pack the source string into the buffer, "
"but only %lu 32-bit words of space are available.  source string = '%s'",
			packed_num_words, maximum_num_words, source );
	}

	if ( num_words_packed != (unsigned long *) NULL ) {
		*num_words_packed = packed_num_words;
	}

	for ( i = 0, j = 0; i < truncated_num_words; i++ ) {

		destination[i] = ((unsigned long) source[j]) << 24;
		j++;
		destination[i] += ((unsigned long) source[j]) << 16;
		j++;
		destination[i] += ((unsigned long) source[j]) << 8;
		j++;
		destination[i] += ((unsigned long) source[j]);
		j++;
	}
	if ( extra_bytes > 0 ) {
		destination[i] = 0;

		for ( k = 0; k < 4; k++ ) {

			if ( j >= string_length ) {
				break;		/* Exit the for() loop. */
			}
			shift = 8 * ( 3 - k );

			destination[i] += ((unsigned long) source[j]) << shift;
			j++;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_mds_display_packed_string( mx_uint32_type *packed_string_buffer,
				unsigned long buffer_length_in_words )
{
	const char fname[] = "mxi_xia_mds_display_packed_string()";

	unsigned long i, this_word;
	char byte0, byte1, byte2, byte3;

	if ( packed_string_buffer == (mx_uint32_type *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The packed string buffer pointer passed was NULL." );
	}

	for ( i = 0; i < buffer_length_in_words; i++ ) {
		this_word = (unsigned long)(packed_string_buffer[i]);

		byte3 = (char) ( (this_word & 0xff000000) >> 24 );
		byte2 = (char) ( (this_word & 0x00ff0000) >> 16 );
		byte1 = (char) ( (this_word & 0x0000ff00) >> 8 );
		byte0 = (char)   (this_word & 0x000000ff);

		MX_DEBUG(-2,("%s: packed string[%lu] = '%c', '%c', '%c', '%c'",
				fname, i, byte3, byte2, byte1, byte0));
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* USE_TCPIP */
