/*
 * Name:    mx_motor.c
 *
 * Purpose: MX function library of generic motor operations.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_key.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_motor.h"

/*=======================================================================*/

/* This function is used by a motor's finish_record_initialization
 * driver entry point to initialize several fields based on the 
 * values in the record description.
 */

MX_EXPORT mx_status_type
mx_motor_finish_record_initialization( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_finish_record_initialization()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *flist;

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}
	motor = (MX_MOTOR *) (motor_record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			motor_record->name );
	}

	/* Initialize scaled quantities from the raw quantities. */

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		motor->position = motor->offset + motor->scale
				* (double) motor->raw_position.stepper;

		motor->backlash_correction = motor->scale
			* (double) motor->raw_backlash_correction.stepper;
		
		motor->negative_limit = motor->offset + motor->scale
				* (double) motor->raw_negative_limit.stepper;

		motor->positive_limit = motor->offset + motor->scale
				* (double) motor->raw_positive_limit.stepper;

		break;
	case MXC_MTR_ANALOG:
		motor->position = motor->offset + motor->scale
					* motor->raw_position.analog;

		motor->backlash_correction = motor->scale
				* motor->raw_backlash_correction.analog;
		
		motor->negative_limit = motor->offset + motor->scale
					* motor->raw_negative_limit.analog;

		motor->positive_limit = motor->offset + motor->scale
					* motor->raw_positive_limit.analog;

		break;
	}

	/* See if the driver can change the motor speed. */

	flist = (MX_MOTOR_FUNCTION_LIST *)
			motor_record->class_specific_function_list;

	if ( flist == (MX_MOTOR_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR_FUNCTION_LIST pointer for record '%s' is NULL.",
			motor_record->name );
	}

	if ( ( flist->get_parameter == NULL )
	  || ( flist->set_parameter == NULL ) )
	{
		motor->motor_flags |= MXF_MTR_CANNOT_CHANGE_SPEED;
		motor->motor_flags |= MXF_MTR_CANNOT_QUICK_SCAN;
	}

	/* Set some plausible initial values for parameters. */

	motor->destination = motor->position;
	motor->set_position = motor->position;

	motor->raw_saved_speed = motor->raw_speed;

	motor->busy = FALSE;
	motor->negative_limit_hit = FALSE;
	motor->positive_limit_hit = FALSE;
	motor->is_at_home_switch = FALSE;

	motor->synchronous_motion_mode = FALSE;

	motor->real_motor_record = NULL;

	/* If 'quick_scan_backlash_correction' is changed to be settable
	 * in the database file, the following will have to be removed.
	 */

	motor->quick_scan_backlash_correction = 0.0;

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_motor_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_motor_get_pointers( MX_RECORD *motor_record,
			MX_MOTOR **motor,
			MX_MOTOR_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_motor_get_pointers()";

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The motor_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( motor_record->mx_class != MXC_MOTOR ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not a motor.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			motor_record->name, calling_fname,
			motor_record->mx_superclass,
			motor_record->mx_class,
			motor_record->mx_type );
	}

	if ( motor != (MX_MOTOR **) NULL ) {
		*motor = (MX_MOTOR *) (motor_record->record_class_struct);

		if ( *motor == (MX_MOTOR *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MOTOR pointer for record '%s' passed by '%s' is NULL",
				motor_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_MOTOR_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_MOTOR_FUNCTION_LIST *)
				(motor_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_MOTOR_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_MOTOR_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				motor_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_motor_is_busy( MX_RECORD *motor_record, int *busy )
{
	const char fname[] = "mx_motor_is_busy()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->motor_is_busy;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"motor_is_busy function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	*busy = motor->busy;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_soft_abort( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_soft_abort()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->soft_abort;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"soft_abort function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_immediate_abort( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_immediate_abort()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR *);
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->immediate_abort;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"immediate_abort function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	return status;
}

/* --- Move by engineering units functions. --- */

MX_EXPORT mx_status_type
mx_motor_move_relative_with_report(MX_RECORD *motor_record,
			double relative_position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags)
{
	double current_position, new_position;
	mx_status_type status;

	status = mx_motor_get_position( motor_record, &current_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	new_position = current_position + relative_position;

	status = mx_motor_move_absolute_with_report( motor_record,
				new_position, move_report_fn, flags );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_move_absolute_with_report(MX_RECORD *motor_record,
			double destination,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags)
{
	const char fname[] = "mx_motor_move_absolute_with_report()";

	MX_MOTOR *motor;
	double raw_destination;
	long raw_motor_steps;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,("%s: moving '%s' to %g",
			fname, motor_record->name, destination ));

	if ( fabs(motor->scale) < MX_MOTOR_ANALOG_FUZZ ) {
		raw_destination = 0.0;
	} else {
		raw_destination = (destination - motor->offset) / motor->scale;
	}

	switch ( motor->subclass ) {
	case MXC_MTR_STEPPER:
		raw_motor_steps = mx_round( raw_destination );

		status = mx_motor_move_absolute_steps_with_report(motor_record,
			raw_motor_steps, move_report_fn, flags );

		break;

	case MXC_MTR_ANALOG:
		status = mx_motor_move_absolute_analog_with_report(motor_record,
			raw_destination, move_report_fn, flags );

		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown motor subclass %d.", motor->subclass );
		break;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_motor_array_move_absolute_with_report( int num_motors,
			MX_RECORD **motor_record_array,
			double *motor_position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags )
{
	const char fname[] = "mx_motor_array_move_absolute_with_report()";

	MX_MOTOR *motor;
	int i, modified_flags, result_flag;
	int do_backlash, individual_backlash_needed;
	double backlash, backlash_position;
	double present_position, relative_motion;
	mx_status_type status;

	/* See if any of the requested moves will exceed a software limit. */

	MX_DEBUG( 2,("%s: Checking software limits.",fname));

	for ( i = 0; i < num_motors; i++ ) {
		status = mx_is_motor_position_between_software_limits(
				motor_record_array[i], motor_position[i],
				&result_flag, TRUE);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	/* Move to the backlash_position. */

	MX_DEBUG( 2,("%s: Checking for backlash correction.", fname));

	if ( (flags & MXF_MTR_IGNORE_BACKLASH) == 0 ) {

		/* We must also check to see if the backlash correction
		 * will exceed a limit.
		 */

		do_backlash = FALSE;

		for ( i = 0; i < num_motors; i++ ) {

			motor = (MX_MOTOR *)
				motor_record_array[i]->record_class_struct;

			backlash = motor->backlash_correction;

			status = mx_motor_get_position( motor_record_array[i],
							&present_position );
			if ( status.code != MXE_SUCCESS )
				return status;

			relative_motion = motor_position[i] - present_position;

			/* A given motor needs to do a backlash correction
			 * if relative_motion and backlash have the same signs.
			 */

			individual_backlash_needed = FALSE;

			if (( relative_motion > 0.0 ) && ( backlash > 0.0 )) {
				individual_backlash_needed = TRUE;
			} else
			if (( relative_motion < 0.0 ) && ( backlash < 0.0 )) {
				individual_backlash_needed = TRUE;
			}

			if ( individual_backlash_needed ) {

				do_backlash = TRUE;

				backlash_position
					= motor_position[i] + backlash;

				status =
				  mx_is_motor_position_between_software_limits(
					motor_record_array[i],
					backlash_position,
					&result_flag, TRUE);

				if ( status.code != MXE_SUCCESS )
					return status;
			}
		}

		/* Do the backlash correction. */

		if ( do_backlash ) {
			modified_flags = flags
					   | MXF_MTR_GO_TO_BACKLASH_POSITION;

			modified_flags &= ( ~ MXF_MTR_NOWAIT );

			MX_DEBUG( 2,
			("%s: Backlash correction for move segment.", fname));

			status = mx_motor_array_move_segment_with_report(
					num_motors,
					motor_record_array,
					motor_position,
					move_report_fn,
					modified_flags );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	}

	/* Now move to the final position. */

	modified_flags = flags | MXF_MTR_IGNORE_BACKLASH;

	MX_DEBUG( 2,("%s: Main move for move segment.", fname));

	status = mx_motor_array_move_segment_with_report(
					num_motors,
					motor_record_array,
					motor_position,
					move_report_fn,
					modified_flags );
	return status;
}

MX_EXPORT mx_status_type
mx_motor_array_move_segment_with_report( int num_motors,
			MX_RECORD **motor_record_array,
			double *motor_position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags )
{
	const char fname[] = "mx_motor_array_move_segment_with_report()";

	MX_RECORD *first_motor_record;
	MX_MOTOR_FUNCTION_LIST *flist;
	mx_status_type (*simultaneous_start_fn)(
					int, MX_RECORD **, double *, int );
	mx_status_type status;
	int move_report_flag, modified_flags;
	int wait_flag;
	int i, j;

	MX_DEBUG( 2,
		("%s invoked.  flags = %d, move_report_fn = %p",
		fname, flags, move_report_fn));

	move_report_flag = flags & MXF_MTR_SHOW_MOVE;

	modified_flags = flags & ( ~ MXF_MTR_SHOW_MOVE );

	modified_flags |= MXF_MTR_NOWAIT;

	if ( flags & MXF_MTR_SIMULTANEOUS_START ) {

		if ( num_motors <= 0 ) {
			/* If there are no motors, then return without
			 * doing anything.
			 */

			return MX_SUCCESSFUL_RESULT;
		}

		first_motor_record = motor_record_array[0];

		/* Get the MX_MOTOR_FUNCTION_LIST for the first motor. */

		flist = (MX_MOTOR_FUNCTION_LIST *)
			first_motor_record->class_specific_function_list;

		if ( flist == (MX_MOTOR_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MOTOR_FUNCTION_LIST pointer for motor '%s' is NULL.",
				first_motor_record->name );
		}

		/* Find the simultaneous_start driver function for the
		 * first motor.
		 */

		simultaneous_start_fn = flist->simultaneous_start;

		if ( simultaneous_start_fn == NULL ) {
			return mx_error( MXE_UNSUPPORTED, fname,
	"The '%s' driver for motor '%s' does not support simultaneous starts.",
				mx_get_driver_name( first_motor_record ),
				first_motor_record->name );
		}

		status = (*simultaneous_start_fn)( num_motors,
						motor_record_array,
						motor_position,
						modified_flags );

		if ( status.code != MXE_SUCCESS )
			return status;
	} else {
		/* We were not asked for a simultaneous start, so start
		 * the motors in sequence.
		 */

		for ( i = 0; i < num_motors; i++ ) {
			MX_DEBUG( 2,("%s: moving %s to %g",
				fname, motor_record_array[i]->name,
				motor_position[i]));

			status = mx_motor_move_absolute_with_report(
				motor_record_array[i], motor_position[i],
				NULL, modified_flags );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < i-1; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j] );
				}
				return status;
			}
		}
	}

	MX_DEBUG( 2,("%s: About to display motor moves.", fname));

	/* Display the motor moves if requested. */

	if ( move_report_flag ) {
		status = ( *move_report_fn ) ( flags,
					num_motors, motor_record_array );
	}

	/* Wait for the motor move to complete if requested. */

	wait_flag = ! ( flags & MXF_MTR_NOWAIT );

	if ( wait_flag ) {
		MX_DEBUG( 2,("%s: wait flag is set.",fname));

		status = mx_wait_for_motor_array_stop( num_motors,
						motor_record_array, flags );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_wait_for_motor_stop( MX_RECORD *motor_record, int flags )
{
	const char fname[] = "mx_wait_for_motor_stop()";

	int busy, interrupt, limit_hit;
	mx_status_type status;

	while (1) {
		status = mx_motor_is_busy(motor_record, &busy);

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( busy == FALSE ) {
			/* Exit the while() loop. */

			break;
		}

		if ( (flags & MXF_MTR_IGNORE_KEYBOARD) == 0 ) {

			/* Did someone hit a key? */

			interrupt = mx_user_requested_interrupt();

			switch( interrupt ) {
			case MXF_USER_INT_NONE:
				/* No interrupt ocurred. */

				break;

			case MXF_USER_INT_ABORT:
				mx_motor_soft_abort( motor_record );

				return mx_error( MXE_INTERRUPTED, fname,
					"Move to backlash position aborted.");

				break;

			case MXF_USER_INT_PAUSE:
				return mx_error( MXE_PAUSE_REQUESTED, fname,
					"Pause requested by user.");

				break;
			default:
				return mx_error( MXE_FUNCTION_FAILED, fname,
					"Unexpected value %d returned "
					"by mx_user_requested_interrupt()",
						    interrupt );
				break;
			}
		}

		/* Has the motor hit a limit? */

		status = mx_motor_positive_limit_hit(
				motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit ) {
			(void) mx_motor_soft_abort( motor_record );

			return mx_error( MXE_INTERRUPTED, fname,
			"Positive limit switch hit.  Motor move aborted." );
		}

		status = mx_motor_negative_limit_hit(
				motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit ) {
			(void) mx_motor_soft_abort( motor_record );

			return mx_error( MXE_INTERRUPTED, fname,
			"Negative limit switch hit.  Motor move aborted." );
		}

		mx_msleep(500);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_wait_for_motor_array_stop( int num_motor_records,
			MX_RECORD **motor_record_array,
			int flags )
{
	const char fname[] = "mx_wait_for_motor_array_stop()";

	int i, j, busy, motor_is_moving;
	int limit_hit, any_limit_hit, interrupt;
	mx_status_type status;

	motor_is_moving = TRUE;
	any_limit_hit = FALSE;

	while ( motor_is_moving && ! any_limit_hit ) {

		motor_is_moving = FALSE;

		for ( i = 0; i < num_motor_records; i++ ) {
			/* Are any motors still moving? */

			status = mx_motor_is_busy(
					motor_record_array[i], &busy );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < num_motor_records; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j]);
				}
				return status;
			}

			if ( busy )
				motor_is_moving = TRUE;

			if ( (flags & MXF_MTR_IGNORE_KEYBOARD) == 0 ){
				/* Did someone hit a key? */

				interrupt = mx_user_requested_interrupt();

				switch( interrupt ) {
				case MXF_USER_INT_NONE:
					/* No interrupt occurred. */

					break;

				case MXF_USER_INT_ABORT:

					for (j=0; j < num_motor_records; j++){
						(void) mx_motor_soft_abort(
							motor_record_array[j]);
					}
					return mx_error(
						MXE_INTERRUPTED, fname,
						"Motor moves aborted.");
					break;

				case MXF_USER_INT_PAUSE:
					return mx_error(
						MXE_PAUSE_REQUESTED,
						fname,
					"Pause requested by user." );

					break;

				default:
					return mx_error(
						MXE_FUNCTION_FAILED,
						fname,
				    "Unexpected value %d returned "
				    "by mx_user_requested_interrupt()",
						interrupt );
					break;
				}
			}

			/* Have any motors hit a limit? */

			status = mx_motor_positive_limit_hit(
				motor_record_array[i], &limit_hit );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < num_motor_records; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j]);
				}
				return status;
			}

			if ( limit_hit )
				any_limit_hit = TRUE;

			status = mx_motor_negative_limit_hit(
				motor_record_array[i], &limit_hit );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < num_motor_records; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j]);
				}
				return status;
			}

			if ( limit_hit )
				any_limit_hit = TRUE;
		}

		if ( any_limit_hit ) {
			for ( j = 0; j < num_motor_records; j++ ) {
				(void) mx_motor_soft_abort(
						motor_record_array[j]);
			}
			return mx_error( MXE_INTERRUPTED, fname,
			"Limit switch hit.  Motor moves aborted." );
		}

		mx_msleep(10);
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_motor_get_position( MX_RECORD *motor_record, double *position )
{
	const char fname[] = "mx_motor_get_position()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	double raw_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"get_position function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		raw_position = (double) ( motor->raw_position.stepper );
		break;
	case MXC_MTR_ANALOG:
		raw_position = motor->raw_position.analog;
		break;
	default:
		raw_position = 0.0;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unrecognized motor subclass %d for motor '%s'.",
			motor->subclass, motor_record->name );
	}

	/* Update position. */

	motor->position = motor->offset + (motor->scale) * raw_position;

	if ( position != NULL ) {
		*position = motor->position;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_position( MX_RECORD *motor_record, double set_position )
{
	const char fname[] = "mx_motor_set_position()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	double raw_set_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"set_position function pointer is NULL.");
	}

	if ( fabs( motor->scale ) < MX_MOTOR_ANALOG_FUZZ ) {
		raw_set_position = 0.0;
	} else {
		raw_set_position
			= ( set_position - motor->offset ) / motor->scale;
	}

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		motor->raw_set_position.stepper = mx_round( raw_set_position );
		break;
	case MXC_MTR_ANALOG:
		motor->raw_set_position.analog = raw_set_position;
		break;
	}

	/* Update set_position. */

	motor->set_position = motor->offset + motor->scale * raw_set_position;

	/* Invoke the function. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_positive_limit_hit( MX_RECORD *motor_record, int *limit_hit )
{
	const char fname[] = "mx_motor_positive_limit_hit()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->positive_limit_hit;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"positive_limit_hit function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	*limit_hit = motor->positive_limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_negative_limit_hit( MX_RECORD *motor_record, int *limit_hit )
{
	const char fname[] = "mx_motor_negative_limit_hit()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->negative_limit_hit;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"negative_limit_hit function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	*limit_hit = motor->negative_limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_find_home_position( MX_RECORD *motor_record, int direction )
{
	const char fname[] = "mx_motor_find_home_position()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->find_home_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"find_home_position function pointer is NULL.");
	}

	if ( motor->scale >= 0.0 ) {
		motor->home_search = direction;
	} else {
		motor->home_search = -direction;
	}

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_zero_position_value( MX_RECORD *motor_record )
{
	mx_status_type status;

	status = mx_motor_set_position( motor_record, 0L );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_constant_velocity_move( MX_RECORD *motor_record, int direction )
{
	const char fname[] = "mx_motor_constant_velocity_move()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->constant_velocity_move;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"constant_velocity_move function pointer is NULL.");
	}

	if ( motor->scale >= 0.0 ) {
		motor->constant_velocity_move = direction;
	} else {
		motor->constant_velocity_move = -direction;
	}

	status = ( *fptr ) ( motor );

	return status;
}

/*-----------------------------------------------------------------------*/

/* Setting a speed limit to -1 means ignore that limit, while setting it
 * to any other negative value means the motor speed may not be changed.
 */

static mx_status_type
mx_motor_check_speed_limits( MX_MOTOR *motor,
				long speed_type,
				double proposed_speed )
{
	const char fname[] = "mx_motor_check_speed_limits()";

	char speed_type_name[20];
	double proposed_raw_speed;
	long test_value;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MOTOR pointer passed was NULL." );
	}

	/* Make sure the engineering limit values match the raw limits. */

	motor->maximum_speed_limit = motor->raw_maximum_speed_limit
					* fabs( motor->scale );

	motor->minimum_speed_limit = motor->raw_minimum_speed_limit
					* fabs( motor->scale );

	proposed_raw_speed = mx_divide_safely( proposed_speed,
						fabs(motor->scale) );

	switch( speed_type ) {
	case MXLV_MTR_SPEED:
		strcpy( speed_type_name, "speed" );
		break;

	case MXLV_MTR_BASE_SPEED:
		strcpy( speed_type_name, "base speed" );
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported speed type %ld was passed for motor '%s'",
			speed_type, motor->record->name );
		break;
	}

	/* Check for negative and zero speeds. */

	if ( proposed_speed <= 0.0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The %s of %g %s/sec requested for motor '%s' is illegal.  "
		"Motor speeds must be positive numbers.",
			speed_type_name, proposed_speed, motor->units,
			motor->record->name );
	}

	/* Check the upper limit. */

	if ( motor->raw_maximum_speed_limit < 0.0 ) {

		test_value = mx_round( motor->raw_maximum_speed_limit );

		/* -1 means ignore the limit. */

		if ( test_value != -1 ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of motor '%s' is not allowed to be changed.",
				speed_type_name, motor->record->name );
		}
	} else {
		if ( proposed_raw_speed > motor->raw_maximum_speed_limit ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of %g %s/sec requested for '%s' is above the "
			"maximum allowed speed of %g %s/sec.",
				speed_type_name, proposed_speed, motor->units,
				motor->record->name,
				motor->maximum_speed_limit, motor->units );
		}
	}

	/* Check the lower limit. */

	if ( motor->raw_minimum_speed_limit < 0.0 ) {

		test_value = mx_round( motor->raw_minimum_speed_limit );

		/* -1 means ignore the limit. */

		if ( test_value != -1 ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of motor '%s' is not allowed to be changed.",
				speed_type_name, motor->record->name );
		}
	} else {
		if ( proposed_raw_speed < motor->raw_minimum_speed_limit ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of %g %s/sec requested for '%s' is below the "
			"minimum allowed speed of %g %s/sec.",
				speed_type_name, proposed_speed, motor->units,
				motor->record->name,
				motor->minimum_speed_limit, motor->units );
		}
	}

	/* Perform speed type specific checks. */

	switch( speed_type ) {
	case MXLV_MTR_SPEED:
		if ( proposed_raw_speed < motor->raw_base_speed ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The %s of %g %s/sec requested for '%s' is below the "
		"current base speed of %g %s/sec.  This is not allowed.",
				speed_type_name, proposed_speed, motor->units,
				motor->record->name,
				motor->base_speed, motor->units );
		}
		break;

	case MXLV_MTR_BASE_SPEED:
		if ( proposed_raw_speed > motor->raw_speed ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The %s of %g %s/sec requested for '%s' is above the "
		"current speed of %g %s/sec.  This is not allowed.",
				speed_type_name, proposed_speed, motor->units,
				motor->record->name,
				motor->speed, motor->units );
		}
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported speed type %ld was passed for motor '%s'",
			speed_type, motor->record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*-----------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mx_motor_get_parameter( MX_RECORD *motor_record, int parameter_type )
{
	const char fname[] = "mx_motor_get_parameter()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The get_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = parameter_type;

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_parameter( MX_RECORD *motor_record, int parameter_type )
{
	const char fname[] = "mx_motor_set_parameter()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The set_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = parameter_type;

	status = ( *fptr ) ( motor );

	return status;
}

/*-----------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mx_motor_default_get_parameter_handler( MX_MOTOR *motor )
{
	const char fname[] = "mx_motor_default_get_parameter_handler()";

	double speed, base_speed;
	double squared_speed, squared_base_speed;
	double acceleration, acceleration_time, acceleration_distance;
	double start_position, end_position;
	double real_start_position, real_end_position;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked for motor '%s', parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string(motor->record,motor->parameter_type),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
	case MXLV_MTR_BASE_SPEED:
	case MXLV_MTR_MAXIMUM_SPEED:
	case MXLV_MTR_ACCELERATION_TYPE:
	case MXLV_MTR_ACCELERATION_PARAMETERS:

		/* These do not require anything to be done. */

		break;

	case MXLV_MTR_ACCELERATION_TIME:

		/* Update the current values of the speed, base speed,
		 * and acceleration parameters.
		 */

		mx_status =
		    mx_motor_update_speed_and_acceleration_parameters( motor );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* We assume the acceleration rate is constant.
		 * 
		 * Remember that the values in motor->acceleration_parameters
		 * are in user units, not raw units.
		 */

		MX_DEBUG( 2,("%s: motor->acceleration_parameters[0] = %g",
				fname, motor->acceleration_parameters[0]));

		MX_DEBUG( 2,("%s: motor->raw_speed = %g",
				fname, motor->raw_speed));

		MX_DEBUG( 2,("%s: motor->raw_base_speed = %g",
				fname, motor->raw_base_speed));

		switch( motor->acceleration_type ) {
		case MXF_MTR_ACCEL_RATE:
			speed = motor->raw_speed * motor->scale;

			base_speed = motor->raw_base_speed * motor->scale;

			acceleration = motor->acceleration_parameters[0];

			acceleration_time = mx_divide_safely(
						speed - base_speed,
						acceleration );

			motor->acceleration_time = fabs( acceleration_time );
			break;

		case MXF_MTR_ACCEL_TIME:
			motor->acceleration_time
				= fabs( motor->acceleration_parameters[0] );

			break;
		default:
			return mx_error( MXE_UNSUPPORTED, fname,
	"Cannot compute the acceleration time for motor '%s' since it has "
	"an unsupported acceleration type of %d", motor->record->name,
						motor->acceleration_type );
			break;
		}

		MX_DEBUG( 2,("%s: motor->acceleration_time = %g",
			fname, motor->acceleration_time));

		break;

	case MXLV_MTR_ACCELERATION_DISTANCE:

		/* Update the current values of the speed, base speed,
		 * and acceleration parameters.
		 */

		mx_status =
		    mx_motor_update_speed_and_acceleration_parameters( motor );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* We assume the acceleration rate is constant.
		 * 
		 * Remember that the values in motor->acceleration_parameters
		 * are in user units, not raw units.
		 */

		switch( motor->acceleration_type ) {
		case MXF_MTR_ACCEL_RATE:
		case MXF_MTR_ACCEL_TIME:
			MX_DEBUG( 2,("%s: motor->raw_speed = %g",
					fname, motor->raw_speed));

			MX_DEBUG( 2,("%s: motor->raw_base_speed = %g",
					fname, motor->raw_base_speed));

			speed = motor->raw_speed * motor->scale;

			base_speed = motor->raw_base_speed * motor->scale;

			MX_DEBUG( 2,("%s: speed = %g, base_speed = %g",
				fname, speed, base_speed ));

			squared_speed = speed * speed;

			MX_DEBUG( 2,("%s: squared_speed = %g",
					fname, squared_speed));

			squared_base_speed = base_speed * base_speed;

			MX_DEBUG( 2,("%s: squared_base_speed = %g",
					fname, squared_base_speed));

			MX_DEBUG( 2,(
				"%s: motor->acceleration_parameters[0] = %g",
				fname, motor->acceleration_parameters[0]));

			if ( motor->acceleration_type == MXF_MTR_ACCEL_RATE )
			{
				MX_DEBUG( 2,("%s: using accel. rate",fname));

				acceleration =
					motor->acceleration_parameters[0];
			} else {
				MX_DEBUG( 2,("%s: using accel. time",fname));

				acceleration = mx_divide_safely(
					speed - base_speed,
					motor->acceleration_parameters[0] );
			}

			acceleration_distance = mx_divide_safely(
					squared_speed - squared_base_speed,
						2.0 * acceleration );

			MX_DEBUG( 2,("%s: acceleration_distance = %g",
				fname, acceleration_distance));

			motor->acceleration_distance
					= fabs( acceleration_distance );
			break;

		default:
			return mx_error( MXE_UNSUPPORTED, fname,
	"Cannot compute the acceleration distance for motor '%s' since it has "
	"an unsupported acceleration type of %d", motor->record->name,
						motor->acceleration_type );
			break;
		}
		break;

	case MXLV_MTR_COMPUTE_EXTENDED_SCAN_RANGE:
		start_position = motor->compute_extended_scan_range[0];
		end_position   = motor->compute_extended_scan_range[1];

		mx_status = mx_motor_get_acceleration_distance( motor->record,
						&acceleration_distance );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( end_position >= start_position ) {

			/* Scanning in the positive direction. */

			real_start_position = start_position
						- acceleration_distance;

			real_end_position = end_position
						+ acceleration_distance;
		} else {

			/* Scanning in the negative direction. */

			real_start_position = start_position
						+ acceleration_distance;

			real_end_position = end_position
						- acceleration_distance;
		}

		motor->compute_extended_scan_range[2] = real_start_position;
		motor->compute_extended_scan_range[3] = real_end_position;

		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
"Parameter type '%s' (%d) is not supported by the MX driver for motor '%s'.",
			mx_get_field_label_string( motor->record,
						motor->parameter_type ),
			motor->parameter_type,
			motor->record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_motor_default_set_parameter_handler( MX_MOTOR *motor )
{
	const char fname[] = "mx_motor_default_set_parameter_handler()";

	double start_position, end_position;
	double time_for_move, requested_speed;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for motor '%s', parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string(motor->record,motor->parameter_type),
		motor->parameter_type));


	switch( motor->parameter_type ) {
	case MXLV_MTR_MAXIMUM_SPEED:
	case MXLV_MTR_ACCELERATION_PARAMETERS:

		/* These do not require anything to be done. */

		break;

	case MXLV_MTR_SPEED:
		if ( motor->raw_speed > motor->raw_maximum_speed ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Requested raw speed %g for motor '%s' is "
				"greater than the raw maximum speed of %g.",
				motor->raw_speed, motor->record->name,
				motor->raw_maximum_speed );
		}
		if ( motor->raw_speed < motor->raw_base_speed ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Requested raw speed %g for motor '%s' is "
				"less than the current raw base speed "
				"of %g.",
				motor->raw_speed, motor->record->name,
				motor->raw_base_speed );
		}
		if ( motor->raw_speed < 0.0 ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Requested raw speed %g for motor '%s' is "
				"less than zero.  Raw motor speeds must be"
				"non-negative numbers.",
				motor->raw_speed, motor->record->name );
		}
		break;
	case MXLV_MTR_BASE_SPEED:
		if ( motor->raw_base_speed > motor->raw_maximum_speed ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Requested raw base speed %g for motor '%s' is "
				"greater than the raw maximum speed of %g.",
				motor->raw_base_speed, motor->record->name,
				motor->raw_maximum_speed );
		}
		if ( motor->raw_base_speed > motor->raw_speed ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Requested raw base speed %g for motor '%s' is "
				"less than the current raw speed of %g.",
				motor->raw_base_speed, motor->record->name,
				motor->raw_speed );
		}
		if ( motor->raw_base_speed < 0.0 ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Requested raw base speed %g for motor '%s' is "
				"less than zero.  Raw motor speeds must be"
				"non-negative numbers.",
				motor->raw_base_speed, motor->record->name );
		}
		break;

	case MXLV_MTR_SPEED_CHOICE_PARAMETERS:
		start_position = motor->speed_choice_parameters[0];
		end_position = motor->speed_choice_parameters[1];
		time_for_move = motor->speed_choice_parameters[2];

		/* We assume the move is at a constant speed in terms
		 * of the current motor.
		 */

		requested_speed = mx_divide_safely(
					end_position - start_position,
					time_for_move );

		requested_speed = fabs( requested_speed );

		MX_DEBUG( 2,("%s: start_position = %g, end_position = %g",
			fname, start_position, end_position));
		MX_DEBUG( 2,("%s: time_for_move = %g, requested_speed = %g",
			fname, time_for_move, requested_speed));

		status = mx_motor_set_speed( motor->record, requested_speed );

		return status;
		break;

	case MXLV_MTR_SAVE_SPEED:
		status = mx_motor_get_parameter( motor->record,
						MXLV_MTR_SPEED );

		if ( status.code != MXE_SUCCESS )
			return status;

		motor->raw_saved_speed = motor->raw_speed;

		MX_DEBUG( 2,("%s (save): motor '%s' raw_saved_speed = %g",
			fname, motor->record->name, motor->raw_saved_speed));
		break;

	case MXLV_MTR_RESTORE_SPEED:
		requested_speed = motor->raw_saved_speed * fabs(motor->scale);

		MX_DEBUG( 2,("%s (restore): motor '%s' raw_saved_speed = %g",
			fname, motor->record->name, motor->raw_saved_speed));
		MX_DEBUG( 2,("%s (restore): motor '%s' scale = %g",
			fname, motor->record->name, motor->scale));
		MX_DEBUG( 2,("%s (restore): motor '%s' requested_speed = %g",
			fname, motor->record->name, requested_speed));

		status = mx_motor_set_speed( motor->record, requested_speed );

		return status;
		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
"Parameter type '%s' (%d) is not supported by the MX driver for motor '%s'.",
			mx_get_field_label_string( motor->record,
						motor->parameter_type ),
			motor->parameter_type,
			motor->record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*-----------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mx_motor_update_speed_and_acceleration_parameters( MX_MOTOR *motor )
{
	mx_status_type mx_status;

	mx_status = mx_motor_get_parameter(motor->record, MXLV_MTR_SPEED);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_motor_get_parameter(motor->record, MXLV_MTR_BASE_SPEED);

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	mx_status = mx_motor_get_parameter(motor->record,
					MXLV_MTR_ACCELERATION_PARAMETERS);

	return mx_status;
}

/*-----------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mx_motor_get_speed( MX_RECORD *motor_record, double *speed )
{
	const char fname[] = "mx_motor_get_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Getting the motor speed is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_SPEED;

	status = ( *fptr ) ( motor );

	motor->speed = fabs(motor->scale) * motor->raw_speed;

	*speed = motor->speed;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_speed( MX_RECORD *motor_record, double speed )
{
	const char fname[] = "mx_motor_set_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Setting the motor speed is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	status = mx_motor_check_speed_limits( motor, MXLV_MTR_SPEED, speed );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the speed. */

	motor->parameter_type = MXLV_MTR_SPEED;

	motor->speed = speed;

	motor->raw_speed = mx_divide_safely(motor->speed, fabs(motor->scale));

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_base_speed( MX_RECORD *motor_record, double *base_speed )
{
	const char fname[] = "mx_motor_get_base_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Getting the motor base speed is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_BASE_SPEED;

	status = ( *fptr ) ( motor );

	motor->base_speed = fabs(motor->scale) * motor->raw_base_speed;

	*base_speed = motor->base_speed;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_base_speed( MX_RECORD *motor_record, double base_speed )
{
	const char fname[] = "mx_motor_set_base_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Setting the motor base speed is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	status = mx_motor_check_speed_limits( motor, MXLV_MTR_BASE_SPEED,
							base_speed );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the base speed. */

	motor->parameter_type = MXLV_MTR_BASE_SPEED;

	motor->base_speed = base_speed;

	motor->raw_base_speed = mx_divide_safely( motor->base_speed,
							fabs(motor->scale) );

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_maximum_speed( MX_RECORD *motor_record, double *maximum_speed )
{
	const char fname[] = "mx_motor_get_maximum_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Getting the motor maximum speed is not supported by "
			"the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_MAXIMUM_SPEED;

	status = ( *fptr ) ( motor );

	motor->maximum_speed = fabs(motor->scale) * motor->raw_maximum_speed;

	*maximum_speed = motor->maximum_speed;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_maximum_speed( MX_RECORD *motor_record, double maximum_speed )
{
	const char fname[] = "mx_motor_set_maximum_speed()";

#if 1
	return mx_error( MXE_UNSUPPORTED, fname,
		"This function is currently disabled." );
#else
	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Setting the motor maximum speed is not supported by "
			"the driver for record '%s'.",
			motor_record->name );
	}

	status = mx_motor_check_speed_limits( motor, MXLV_MTR_MAXIMUM_SPEED,
							maximum_speed );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the maximum speed. */

	motor->parameter_type = MXLV_MTR_MAXIMUM_SPEED;

	motor->maximum_speed = maximum_speed;

	motor->raw_maximum_speed = mx_divide_safely( motor->maximum_speed,
							fabs(motor->scale) );

	status = ( *fptr ) ( motor );

	return status;
#endif
}

/*
 * mx_motor_save_speed() saves both the speed and the value of
 * motor->synchronous_motion_mode.
 */

MX_EXPORT mx_status_type
mx_motor_save_speed( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_save_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Saving the motor speed is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->saved_synchronous_motion_mode = motor->synchronous_motion_mode;

	motor->parameter_type = MXLV_MTR_SAVE_SPEED;

	status = ( *fptr ) ( motor );

	return status;
}

/*
 * mx_motor_restore_speed() restores both the speed and the value of
 * motor->synchronous_motion_mode.
 */

MX_EXPORT mx_status_type
mx_motor_restore_speed( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_restore_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Restoring the motor speed is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_RESTORE_SPEED;

	status = ( *fptr ) ( motor );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor->synchronous_motion_mode = motor->saved_synchronous_motion_mode;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_motor_get_synchronous_motion_mode( MX_RECORD *motor_record,
					int *synchronous_motion_mode )
{
	const char fname[] = "mx_motor_get_synchronous_motion_mode()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Getting the synchronous motion mode is not supported "
			"by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_SYNCHRONOUS_MOTION_MODE;

	status = ( *fptr ) ( motor );

	if ( synchronous_motion_mode != NULL ) {
		*synchronous_motion_mode = motor->synchronous_motion_mode;
	}
	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_synchronous_motion_mode( MX_RECORD *motor_record,
					int synchronous_motion_mode )
{
	const char fname[] = "mx_motor_set_synchronous_motion_mode()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Setting the synchronous motion mode is not supported "
			"by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_SYNCHRONOUS_MOTION_MODE;

	motor->synchronous_motion_mode = synchronous_motion_mode;

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_speed_between_positions( MX_RECORD *motor_record,
					double position1, double position2,
					double time_for_move )
{
	const char fname[] = "mx_motor_set_speed_between_positions()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Setting the motor speed between two positions is not "
		"supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->speed_choice_parameters[0] = position1;
	motor->speed_choice_parameters[1] = position2;
	motor->speed_choice_parameters[2] = time_for_move;

	motor->parameter_type = MXLV_MTR_SPEED_CHOICE_PARAMETERS;

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_acceleration_type( MX_RECORD *motor_record,
				int *acceleration_type )
{
	const char fname[] = "mx_motor_get_acceleration_type()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Getting the acceleration type is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_TYPE;

	status = ( *fptr ) ( motor );

	*acceleration_type = motor->acceleration_type;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_acceleration_parameters( MX_RECORD *motor_record,
				double *acceleration_parameter_array )
{
	const char fname[] = "mx_motor_get_acceleration_parameters()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;
	int i;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Getting acceleration parameters is not supported by "
			"the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_PARAMETERS;

	status = ( *fptr ) ( motor );

	for ( i = 0; i < MX_MOTOR_NUM_ACCELERATION_PARAMS; i++ ) {
		acceleration_parameter_array[i] =
				motor->acceleration_parameters[i];
	}

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_acceleration_parameters( MX_RECORD *motor_record,
				double *acceleration_parameter_array )
{
	const char fname[] = "mx_motor_set_acceleration_parameters()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;
	int i;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Setting acceleration parameters is not supported "
			"by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_PARAMETERS;

	for ( i = 0; i < MX_MOTOR_NUM_ACCELERATION_PARAMS; i++ ) {
		motor->acceleration_parameters[i] =
				acceleration_parameter_array[i];
	}

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_acceleration_time( MX_RECORD *motor_record,
				double *acceleration_time )
{
	const char fname[] = "mx_motor_get_acceleration_time()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"Getting the acceleration time is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_TIME;

	status = ( *fptr ) ( motor );

	*acceleration_time = motor->acceleration_time;

	MX_DEBUG( 2,("%s: motor '%s' acceleration time = %g",
		fname, motor_record->name, *acceleration_time));

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_acceleration_distance( MX_RECORD *motor_record,
				double *acceleration_distance )
{
	const char fname[] = "mx_motor_get_acceleration_distance()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Getting the acceleration distance is not supported "
			"by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_DISTANCE;

	status = ( *fptr ) ( motor );

	*acceleration_distance = motor->acceleration_distance;

	MX_DEBUG( 2,("%s: motor '%s' acceleration distance = %g",
		fname, motor_record->name, *acceleration_distance));

	return status;
}

MX_EXPORT mx_status_type
mx_motor_is_at_home_switch( MX_RECORD *motor_record, int *is_at_home_switch )
{
	const char fname[] = "mx_motor_is_at_home_switch()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Checking to see if a motor is at the home switch is not "
		"supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_IS_AT_HOME_SWITCH;

	status = ( *fptr ) ( motor );

	*is_at_home_switch = motor->is_at_home_switch;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_compute_extended_scan_range( MX_RECORD *motor_record,
					double start_position,
					double end_position,
					double *real_start_position,
					double *real_end_position )
{
	const char fname[] = "mx_motor_compute_extended_scan_range()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Computing an extended scan range is not supported "
			"by the driver for record '%s'.",
			motor_record->name );
	}

	motor->compute_extended_scan_range[0] = start_position;
	motor->compute_extended_scan_range[1] = end_position;

	motor->parameter_type = MXLV_MTR_COMPUTE_EXTENDED_SCAN_RANGE;

	status = ( *fptr ) ( motor );

	*real_start_position = motor->compute_extended_scan_range[2];
	*real_end_position = motor->compute_extended_scan_range[3];

	return status;
}

MX_EXPORT mx_status_type
mx_motor_compute_pseudomotor_position_from_real_position(
					MX_RECORD *motor_record,
					double real_position,
					double *pseudomotor_position,
					int recursion_flag )
{
	const char fname[] =
		"mx_motor_compute_pseudomotor_position_from_real_position()";

	static int level = 0;

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	MX_RECORD *next_level_motor_record;
	double next_level_pseudomotor_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	level++;

	MX_DEBUG( 2,
	("%s(%d): motor '%s', real_position = %g, recursion_flag = %d",
	fname, level, motor_record->name, real_position, recursion_flag ));

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		level--;

		return mx_error( MXE_UNSUPPORTED, fname,
			"Computing a pseudomotor position from a real position "
			"is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	/* First, perform special processing depending on what flags are set
	 * for this motor record.
	 */

	if ( motor->motor_flags & MXF_MTR_IS_PSEUDOMOTOR ) {

		/* If this is a pseudomotor and recursion is turned on,
		 * we must recurse down to the next lower layer of pseudo-
		 * motor.  We find the next level motor record by using
		 * the motor->real_motor_record pointer.
		 */

		if ( recursion_flag ) {
			next_level_motor_record = motor->real_motor_record;

			if ( next_level_motor_record == (MX_RECORD *)NULL ) {
				level--;

				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Cannot compute a unique pseudomotor position for motor '%s' since this type "
"of pseudomotor does not support it.", motor_record->name );
			}
			status =
		    mx_motor_compute_pseudomotor_position_from_real_position(
				next_level_motor_record,
				real_position,
				&next_level_pseudomotor_position,
				recursion_flag );

			if ( status.code != MXE_SUCCESS ) {
				level--;

				return status;
			}

			motor->compute_pseudomotor_position[0]
					= next_level_pseudomotor_position;
		} else {
			motor->compute_pseudomotor_position[0] = real_position;
		}

	} else if ( motor->motor_flags & MXF_MTR_IS_REMOTE_MOTOR ) {

		motor->compute_pseudomotor_position[0] = real_position;

	} else {
		/* If this is not a pseudomotor or a remote motor, we assume
		 * that this is a real motor and return without performing
		 * any transformation on the position.
		 */

		*pseudomotor_position = real_position;

		motor->compute_pseudomotor_position[1] = real_position;

		MX_DEBUG( 2,("%s(%d,real): *pseudomotor_position = %g",
			fname, level, *pseudomotor_position ));

		level--;

		return MX_SUCCESSFUL_RESULT;
	}

	/* Perform this level of pseudomotor (or remote motor) conversion. */

	motor->compute_pseudomotor_position[1] = 0.0;

	motor->parameter_type = MXLV_MTR_COMPUTE_PSEUDOMOTOR_POSITION;

	status = ( *fptr ) ( motor );

	*pseudomotor_position = motor->compute_pseudomotor_position[1];

	if ( motor->motor_flags & MXF_MTR_IS_REMOTE_MOTOR ) {
		MX_DEBUG( 2,("%s(%d,remote): *pseudomotor_position = %g",
			fname, level, *pseudomotor_position ));
	} else {
		MX_DEBUG( 2,("%s(%d,pseudo): *pseudomotor_position = %g",
			fname, level, *pseudomotor_position ));
	}

	level--;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_compute_real_position_from_pseudomotor_position(
					MX_RECORD *motor_record,
					double pseudomotor_position,
					double *real_position,
					int recursion_flag )
{
	const char fname[] =
		"mx_motor_compute_real_position_from_pseudomotor_position()";

	static int level = 0;

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	MX_RECORD *next_level_motor_record;
	double next_level_real_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	level++;

	MX_DEBUG( 2,
	("%s(%d): motor '%s', pseudomotor_position = %g, recursion_flag = %d",
		fname, level, motor_record->name, pseudomotor_position,
		recursion_flag ));

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		level--;

		return mx_error( MXE_UNSUPPORTED, fname,
		"Computing a real motor position from a pseudomotor "
		"position is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->compute_real_position[0] = pseudomotor_position;
	motor->compute_real_position[1] = 0.0;

	/* If this motor is not a pseudomotor or a remote motor, we can
	 * return immediately without performing any transformations
	 * on the position.
	 */

	if ( ((motor->motor_flags & MXF_MTR_IS_PSEUDOMOTOR) == 0)
	  && ((motor->motor_flags & MXF_MTR_IS_REMOTE_MOTOR) == 0) )
	{
		*real_position = pseudomotor_position;

		motor->compute_real_position[1] = pseudomotor_position;

		MX_DEBUG( 2,("%s(%d,real): *real_position = %g",
			fname, level, *real_position ));

		level--;

		return MX_SUCCESSFUL_RESULT;
	}

	/* Perform this level of pseudomotor (or remote motor) conversion. */

	motor->parameter_type = MXLV_MTR_COMPUTE_REAL_POSITION;

	status = ( *fptr ) ( motor );

	if ( status.code != MXE_SUCCESS ) {
		level--;
		return status;
	}

	*real_position = motor->compute_real_position[1];

	/* If this is a remote motor, no further transformation is
	 * necessary and we return.
	 */

	if ( motor->motor_flags & MXF_MTR_IS_REMOTE_MOTOR ) {
		level--;

		MX_DEBUG( 2,("%s(%d,remote): *real_position = %g",
			fname, level, *real_position));

		return MX_SUCCESSFUL_RESULT;
	}

	/* If recursion is turned on, we must recurse down to the next
	 * lower layer of pseudomotor.
	 */

	if ( recursion_flag ) {

		next_level_motor_record = motor->real_motor_record;

		if ( next_level_motor_record == (MX_RECORD *) NULL ) {
			level--;

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Cannot compute a unique real position for motor '%s' since this type "
"of pseudomotor does not support it.", motor_record->name );
		}

		status
		  = mx_motor_compute_real_position_from_pseudomotor_position(
				next_level_motor_record,
				*real_position,
				&next_level_real_position,
				recursion_flag );

		if ( status.code != MXE_SUCCESS ) {
			level--;
			return status;
		}

		*real_position = next_level_real_position;
	}

	motor->compute_real_position[1] = *real_position;

	MX_DEBUG( 2,("%s(%d,pseudo): *real_position = %g",
			fname, level, *real_position));

	level--;

	return MX_SUCCESSFUL_RESULT;
}

/* === Move by steps functions. (MXC_MTR_STEPPER) === */

MX_EXPORT mx_status_type
mx_motor_move_relative_steps_with_report(MX_RECORD *motor_record,
			long relative_steps,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags)
{
	long current_position, new_position;
	mx_status_type status;

	status = mx_motor_get_position_steps(motor_record, &current_position);

	if ( status.code != MXE_SUCCESS )
		return status;

	new_position = current_position + relative_steps;

	status = mx_motor_move_absolute_steps_with_report( motor_record,
			new_position, move_report_fn, flags );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_move_absolute_steps_with_report(MX_RECORD *motor_record,
			long motor_steps,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags)
{
	const char fname[] = "mx_motor_move_absolute_steps_with_report()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	long positive_limit, negative_limit;
	long position, backlash;
	long relative_motion, backlash_position;
	double dummy_position;
	int mask, do_backlash;
	mx_status_type status, move_report_status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	fptr = fl_ptr->move_absolute;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"move_absolute function pointer is NULL.");
	}

	positive_limit = motor->raw_positive_limit.stepper;
	negative_limit = motor->raw_negative_limit.stepper;

	status = mx_motor_get_position_steps(motor_record, &position);

	if ( status.code != MXE_SUCCESS )
		return status;

	relative_motion = motor_steps - position;

	/* Do not do the move if it would be less than the deadband size. */

	if ( labs( relative_motion ) < motor->deadband.stepper ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Do we need to do a backlash correction this time? */

	backlash = motor->raw_backlash_correction.stepper;

	if ( flags & MXF_MTR_IGNORE_BACKLASH ) {
		do_backlash = FALSE;
	} else {
		do_backlash = TRUE;
	}

	if ( do_backlash ) {
		/* Perform the backlash if relative_motion and backlash
		 * are both non-zero and have the same signs.
		 */

		do_backlash = FALSE;

		if ( ( relative_motion > 0 ) && ( backlash > 0 ) ) {
			do_backlash = TRUE;
		} else if ( ( relative_motion < 0 ) && ( backlash < 0 ) ) {
			do_backlash = TRUE;
		}
	}

	/* Do the backlash correction, if necessary. */

	if ( do_backlash ) {
		backlash_position = motor_steps + backlash;

		/* Check software limits. */

		if ( backlash_position > positive_limit ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Backlash correction for '%s' to %ld steps would exceed positive limit "
	"at %ld steps.",
				motor_record->name,
				backlash_position,
				positive_limit);
		}

		if ( backlash_position < negative_limit ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Backlash correction for '%s' to %ld steps would exceed negative limit "
	"at %ld steps.",
				motor_record->name,
				backlash_position,
				negative_limit);
		}

		/* Do not do the backlash correction if we were only asked
		 * to check limits.
		 */

		if ( (flags & MXF_MTR_ONLY_CHECK_LIMITS) == 0 ) {

			/* Update destination. */

			motor->destination = motor->offset
				+ motor->scale * (double) backlash_position;

			/* Do the move. */

			motor->raw_destination.stepper = backlash_position;

			status = ( *fptr ) ( motor );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Leave now if we were not asked to wait for the
			 * completion of the backlash correction.
			 */

			mask = MXF_MTR_NOWAIT | MXF_MTR_GO_TO_BACKLASH_POSITION;

			if ( (flags & mask) == mask ) {
				return MX_SUCCESSFUL_RESULT;
			}

			/* Otherwise, display our progress on the way to
			 * the backlash position.
			 */

			if ( move_report_fn != NULL ) {
				move_report_status = (*move_report_fn)
						( flags, 1, &motor_record );
			} else {
				move_report_status = MX_SUCCESSFUL_RESULT;

				status = mx_wait_for_motor_stop(
						motor_record, flags );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			/* Update copy of motor position in MOTOR structure. */

			status = mx_motor_get_position(
					motor_record, &dummy_position);

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Return if move was aborted. */

			if ( move_report_status.code != MXE_SUCCESS )
				return move_report_status;
		}
	}

	/* Were we asked to only go to the backlash position? */

	if ( flags & MXF_MTR_GO_TO_BACKLASH_POSITION ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Check software limits. */

	if ( motor_steps > positive_limit ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Error: move of '%s' to %ld steps would exceed positive limit at %ld steps.",
			motor_record->name,
			motor_steps,
			positive_limit );
	}

	if ( motor_steps < negative_limit ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Error: move of '%s' to %ld steps would exceed negative limit at %ld steps.",
			motor_record->name,
			motor_steps,
			negative_limit );
	}

	/* Return now if we are only checking limits. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Update the destination. */

	motor->destination = motor->offset
				+ motor->scale * (double) motor_steps;

	/* Do the move. */

	motor->raw_destination.stepper = motor_steps;

	status = ( *fptr ) ( motor );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( move_report_fn != NULL ) {
		move_report_status = move_report_fn(
						flags, 1, &motor_record);
	} else {
		move_report_status = MX_SUCCESSFUL_RESULT;
	}

	/* Update copy of motor position in MX_MOTOR structure, unless
	 * MXF_MTR_NOWAIT is set in 'flags'.
	 */

	if ( (flags & MXF_MTR_NOWAIT) == 0 ) {
		status = mx_motor_get_position(motor_record, &dummy_position);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return move_report_status;
}

MX_EXPORT mx_status_type
mx_motor_get_position_steps( MX_RECORD *motor_record, long *motor_steps )
{
	const char fname[] = "mx_motor_get_position_steps()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	fptr = fl_ptr->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"get_position_steps function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	/* Update position. */

	motor->position = motor->offset
			+ (motor->scale) * motor->raw_position.stepper;

	*motor_steps = motor->raw_position.stepper;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_position_steps( MX_RECORD *motor_record, long motor_steps )
{
	const char fname[] = "mx_motor_set_position_steps()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	fptr = fl_ptr->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"set_position_steps function pointer is NULL.");
	}

	/* Update set_position. */

	motor->raw_set_position.stepper = motor_steps;

	motor->set_position = motor->offset
			+ motor->scale * (double) motor_steps;

	/* Invoke the function. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_steps_to_go( MX_RECORD *motor_record, long *steps_to_go )
{
	const char fname[] = "mx_motor_steps_to_go()";

	MX_MOTOR *motor;
	long step_position, step_destination;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	status = mx_motor_get_position_steps( motor_record, &step_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	step_destination = motor->raw_destination.stepper;

	*steps_to_go = step_destination - step_position;

	return status;
}

/* === Move analog functions (MXC_MTR_ANALOG) === */

MX_EXPORT mx_status_type
mx_motor_move_absolute_analog_with_report(MX_RECORD *motor_record,
			double motor_position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags)
{
	const char fname[] = "mx_motor_move_absolute_analog_with_report()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	double positive_limit, negative_limit;
	double present_position, backlash;
	double relative_motion, backlash_position;
	double test_var, dummy_position;
	int mask, do_backlash;
	char units[80];
	mx_status_type status, move_report_status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor->record->name );
	}

	fptr = fl_ptr->move_absolute;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"move_absolute function pointer is NULL.");
	}

	positive_limit = motor->raw_positive_limit.analog;
	negative_limit = motor->raw_negative_limit.analog;

	status = mx_motor_get_position_analog(motor_record, &present_position);

	if ( status.code != MXE_SUCCESS )
		return status;

	relative_motion = motor_position - present_position;

	/* Do not do the move if it would be less than the deadband size. */

	if ( fabs( relative_motion ) < motor->deadband.analog ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Do we need to do a backlash correction this time? */

	backlash = motor->raw_backlash_correction.analog;

	if ( flags & MXF_MTR_IGNORE_BACKLASH ) {
		do_backlash = FALSE;
	} else {
		do_backlash = TRUE;
	}

	if ( do_backlash ) {
		/* Perform the backlash if relative_motion and backlash
		 * are both non-zero and have the same signs.
		 */

		do_backlash = FALSE;

		if ( ( relative_motion > 0.0 ) && ( backlash > 0.0 ) ) {
			do_backlash = TRUE;
		} else if ( ( relative_motion < 0.0 ) && ( backlash < 0.0 ) ) {
			do_backlash = TRUE;
		}
	}

	/* Do the backlash correction, if necessary. */

	/* MX currently doesn't record the raw units for analog motors.
	 * If the absolute value of the scale factor for this motor is
	 * very close to 1, I can assume that the raw units are the 
	 * same as the user units.  Otherwise, I don't know what the
	 * units are.  This is a kludge until I replace this with 
	 * something better.
	 */

	if ( do_backlash ) {
		backlash_position = motor_position + backlash;

		/* Check software limits. */

		if ( backlash_position > positive_limit ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
							sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Backlash correction for '%s' to %.*g%s would exceed positive limit at %.*g%s.",
				motor->record->name,
				motor->record->precision,
				backlash_position, units,
				motor->record->precision,
				positive_limit, units );
		}

		if ( backlash_position < negative_limit ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
							sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Backlash correction for '%s' to %.*g%s would exceed negative limit at %.*g%s.",
				motor->record->name,
				motor->record->precision,
				backlash_position, units,
				motor->record->precision,
				negative_limit, units );
		}

		/* Do not do the backlash correction if we were only asked
		 * to check limits.
		 */

		if ( (flags & MXF_MTR_ONLY_CHECK_LIMITS) == 0 ) {

			/* Update the destination. */

			motor->destination = motor->offset
				+ motor->scale * backlash_position;

			/* Do the move. */

			motor->raw_destination.analog = backlash_position;

			status = ( *fptr ) ( motor );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Leave now if we were not asked to wait for the
			 * completion of the backlash correction.
			 */

			mask = MXF_MTR_NOWAIT | MXF_MTR_GO_TO_BACKLASH_POSITION;

			if ( (flags & mask) == mask ) {
				return MX_SUCCESSFUL_RESULT;
			}

			/* Otherwise, display our progress on the way to
			 * the backlash position.
			 */

			if ( move_report_fn != NULL ) {
				move_report_status = (*move_report_fn)
						( flags, 1, &motor_record );
			} else {
				move_report_status = MX_SUCCESSFUL_RESULT;

				status = mx_wait_for_motor_stop(
						motor_record, flags );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			/* Update copy of motor position in MOTOR structure. */

			status = mx_motor_get_position(
					motor_record, &dummy_position);

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Return if move was aborted. */

			if ( move_report_status.code != MXE_SUCCESS )
				return move_report_status;
		}
	}

	/* Were we asked to only go to the backlash position? */

	if ( flags & MXF_MTR_GO_TO_BACKLASH_POSITION ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Check software limits. */

	if ( motor_position > positive_limit ) {
		test_var = fabs( fabs( motor->scale ) - 1.0 );

		if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
			strcpy( units, " " );
			strncat( units, motor->units, sizeof(units) );
		} else {
			strcpy( units, "" );
		}

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Error: move of '%s' to %.*g%s would exceed positive limit at %.*g%s.",
			motor->record->name,
			motor->record->precision,
			motor_position, units,
			motor->record->precision,
			positive_limit, units );
	}

	if ( motor_position < negative_limit ) {
		test_var = fabs( fabs( motor->scale ) - 1.0 );

		if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
			strcpy( units, " " );
			strncat( units, motor->units, sizeof(units) );
		} else {
			strcpy( units, "" );
		}

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Error: move of '%s' to %.*g%s would exceed negative limit at %.*g%s.",
			motor->record->name,
			motor->record->precision,
			motor_position, units,
			motor->record->precision,
			negative_limit, units );
	}

	/* Return now if we are only checking limits. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Update the destination. */

	motor->destination = motor->offset
				+ motor->scale * motor_position;

	/* Do the move. */

	motor->raw_destination.analog = motor_position;

	status = ( *fptr ) ( motor );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( move_report_fn != NULL ) {
		move_report_status = move_report_fn(
						flags, 1, &motor_record);
	} else {
		move_report_status = MX_SUCCESSFUL_RESULT;
	}

	/* Update copy of motor position in MX_MOTOR structure, unless
	 * MXF_MTR_NOWAIT is set in 'flags'.
	 */

	if ( (flags & MXF_MTR_NOWAIT) == 0 ) {
		status = mx_motor_get_position(motor_record, &dummy_position);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return move_report_status;
}

MX_EXPORT mx_status_type
mx_motor_get_position_analog( MX_RECORD *motor_record,
				double *raw_position )
{
	const char fname[] = "mx_motor_get_position_analog()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor_record->name );
	}

	fptr = fl_ptr->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"get_position_analog function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	/* Update position. */

	motor->position = motor->offset
			+ (motor->scale) * motor->raw_position.analog;

	*raw_position = motor->raw_position.analog;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_position_analog( MX_RECORD *motor_record,
						double raw_set_position )
{
	const char fname[] = "mx_motor_set_position_analog()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor_record->name );
	}

	fptr = fl_ptr->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"set_position_analog function pointer is NULL.");
	}

	/* Update set_position. */

	motor->raw_set_position.analog = raw_set_position;

	motor->set_position = motor->offset + motor->scale * raw_set_position;

	/* Invoke the function. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_is_motor_position_between_software_limits(
	MX_RECORD *motor_record, double proposed_position,
	int *result_flag, int generate_error_message )
{
	const char fname[] =
		"mx_is_motor_position_between_software_limits()";

	MX_MOTOR *motor;
	double scale, offset, unscaled_proposed_position;
	long proposed_step_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	scale = motor->scale;
	offset = motor->offset;

	if ( fabs(scale) < MX_MOTOR_ANALOG_FUZZ ) {
		unscaled_proposed_position = 0;
	} else {
		unscaled_proposed_position
			= (proposed_position - offset) / scale;
	}

	switch( motor->subclass ) {
	case MXC_MTR_ANALOG:
	    status =
		mx_is_analog_motor_position_between_software_limits(
			motor_record, unscaled_proposed_position,
			result_flag, generate_error_message );
	    break;
	case MXC_MTR_STEPPER:
	    proposed_step_position = mx_round( unscaled_proposed_position );

	    status =
		mx_is_stepper_motor_position_between_software_limits(
			motor_record, proposed_step_position,
			result_flag, generate_error_message );
	    break;
	default:
	    return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Unrecognized motor subclass = %d", motor->subclass );
	    break;
	}
	return status;
}

MX_EXPORT mx_status_type
mx_is_analog_motor_position_between_software_limits(
	MX_RECORD *motor_record, double proposed_analog_position,
	int *result_flag, int generate_error_message )
{
	const char fname[] =
	    "mx_is_analog_motor_position_between_software_limits()";

	MX_MOTOR *motor;
	char message_buffer[120];
	char units[80];
	double positive_limit, negative_limit, test_var;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor->record->name );
	}

	if ( result_flag == (int *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"result_flag pointer passed was NULL." );
	}

	message_buffer[0] = '\0';
	*result_flag = 0;

	positive_limit = motor->raw_positive_limit.analog;
	negative_limit = motor->raw_negative_limit.analog;

	/* MX currently doesn't record the raw units for analog motors.
	 * If the absolute value of the scale factor for this motor is
	 * very close to 1, I can assume that the raw units are the 
	 * same as the user units.  Otherwise, I don't know what the
	 * units are.  This is a kludge until I replace this with 
	 * something better.
	 */

	if ( proposed_analog_position > positive_limit ) {
		*result_flag = 1;

		if ( generate_error_message ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
						sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			sprintf( message_buffer,
		"Move of '%s' to %.*g%s would exceed positive limit at %.*g%s.",
				motor_record->name,
				motor_record->precision,
				proposed_analog_position, units,
				motor_record->precision,
				positive_limit, units );
		}
	}
	if ( proposed_analog_position < negative_limit ) {
		*result_flag = -1;

		if ( generate_error_message ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
						sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			sprintf( message_buffer,
		"Move of '%s' to %.*g%s would exceed negative limit at %.*g%s.",
				motor_record->name,
				motor_record->precision,
				proposed_analog_position, units,
				motor_record->precision,
				negative_limit, units );
		}
	}

	if ( (*result_flag != 0) && generate_error_message ) {
		return mx_error(MXE_WOULD_EXCEED_LIMIT, fname, message_buffer);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_is_stepper_motor_position_between_software_limits(
	MX_RECORD *motor_record, long proposed_step_position,
	int *result_flag, int generate_error_message )
{
	const char fname[] =
	    "mx_is_stepper_motor_position_between_software_limits()";

	MX_MOTOR *motor;
	char message_buffer[120];
	long positive_limit, negative_limit;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an stepper motor.",
			motor->record->name );
	}

	if ( result_flag == (int *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"result_flag pointer passed was NULL." );
	}

	message_buffer[0] = '\0';
	*result_flag = 0;

	positive_limit = motor->raw_positive_limit.stepper;
	negative_limit = motor->raw_negative_limit.stepper;

	if ( proposed_step_position > positive_limit ) {
		*result_flag = 1;

		if ( generate_error_message ) {
			sprintf( message_buffer,
	"Move of '%s' to %ld steps would exceed positive limit at %ld steps.",
				motor_record->name,
				proposed_step_position,
				positive_limit );
		}
	}
	if ( proposed_step_position < negative_limit ) {
		*result_flag = -1;

		if ( generate_error_message ) {
			sprintf( message_buffer,
	"Move of '%s' to %ld steps would exceed negative limit at %ld steps.",
				motor_record->name,
				proposed_step_position,
				negative_limit );
		}
	}

	if ( (*result_flag != 0) && generate_error_message ) {
		return mx_error(MXE_WOULD_EXCEED_LIMIT, fname, message_buffer);
	}

	return MX_SUCCESSFUL_RESULT;
}

