/*
 * Name:    v_old_mathop.c
 *
 * Purpose: Support for simple mathematical operations on a list of records.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "mx_record.h"
#include "mx_driver.h"
#include "mx_types.h"
#include "mx_variable.h"
#include "v_old_mathop.h"

#include "mx_analog_input.h"
#include "mx_analog_output.h"
#include "mx_digital_input.h"
#include "mx_digital_output.h"
#include "mx_motor.h"
#include "mx_encoder.h"
#include "mx_scaler.h"
#include "mx_timer.h"
#include "mx_amplifier.h"
#include "mx_relay.h"

MX_RECORD_FUNCTION_LIST mxv_old_mathop_record_function_list = {
	mxv_old_mathop_initialize_type,
	mxv_old_mathop_create_record_structures,
	mxv_old_mathop_finish_record_initialization,
	mx_default_delete_record_handler,
	NULL,
	mxv_old_mathop_dummy_function,
	mxv_old_mathop_dummy_function,
	mxv_old_mathop_dummy_function,
	mxv_old_mathop_dummy_function
};

MX_VARIABLE_FUNCTION_LIST mxv_old_mathop_variable_function_list = {
	mxv_old_mathop_send_variable,
	mxv_old_mathop_receive_variable
};

MX_RECORD_FIELD_DEFAULTS mxv_old_mathop_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_OLD_MATHOP_VARIABLE_STANDARD_FIELDS,
	MX_VARIABLE_STANDARD_FIELDS,
	MX_DOUBLE_VARIABLE_STANDARD_FIELDS
};

long mxv_old_mathop_num_record_fields
	= sizeof( mxv_old_mathop_record_field_defaults )
	/ sizeof( mxv_old_mathop_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxv_old_mathop_rfield_def_ptr
		= &mxv_old_mathop_record_field_defaults[0];

#define MXV_OLD_MATHOP_DEBUG	FALSE

/********************************************************************/

static mx_status_type mxv_old_mathop_get_value( MX_RECORD *record, double *value );

static mx_status_type mxv_old_mathop_put_value( MX_RECORD *record, double value );

static mx_status_type mxv_old_mathop_compute_value( MX_VARIABLE *variable,
						double *value,
						MX_RECORD **destination_record);

MX_EXPORT mx_status_type
mxv_old_mathop_initialize_type( long record_type )
{
	const char fname[] = "mxv_old_mathop_initialize_type()";

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	long num_record_fields, referenced_field_index;
	long num_records_varargs_cookie;
	mx_status_type mx_status;

	mx_status = mx_variable_initialize_type( record_type );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr = driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	num_record_fields = *(driver->num_record_fields);

	/* Construct a varargs cookie for 'num_records'. */

	mx_status = mx_find_record_field_defaults_index(
			record_field_defaults, num_record_fields,
			"num_records", &referenced_field_index );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_construct_varargs_cookie( referenced_field_index, 0,
					&num_records_varargs_cookie );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Initialize all the varargs fields that depend on 'num_records'. */

	mx_status = mx_find_record_field_defaults( record_field_defaults,
				num_record_fields, "record_name_array", &field);

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	field->dimension[0] = num_records_varargs_cookie;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_old_mathop_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxv_old_mathop_create_record_structures()";

	MX_VARIABLE *variable_struct;
	MX_OLD_MATHOP_VARIABLE *old_mathop_variable;

	/* Allocate memory for the necessary structures. */

	variable_struct = (MX_VARIABLE *) malloc( sizeof(MX_VARIABLE) );

	if ( variable_struct == (MX_VARIABLE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_VARIABLE structure." );
	}

	old_mathop_variable = (MX_OLD_MATHOP_VARIABLE *)
				malloc( sizeof(MX_OLD_MATHOP_VARIABLE) );

	if ( old_mathop_variable == (MX_OLD_MATHOP_VARIABLE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_OLD_MATHOP_VARIABLE structure." );
	}

	/* Now set up the necessary pointers. */

	variable_struct->record = record;

	record->record_superclass_struct = variable_struct;
	record->record_class_struct = NULL;
	record->record_type_struct = old_mathop_variable;

	record->superclass_specific_function_list =
				&mxv_old_mathop_variable_function_list;
	record->class_specific_function_list = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_old_mathop_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxv_old_mathop_finish_record_initialization()";

	MX_OLD_MATHOP_VARIABLE *old_mathop_variable;
	MX_RECORD *dependent_record;
	char *name, *endptr;
	long i, num_records;
	double value;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	old_mathop_variable = (MX_OLD_MATHOP_VARIABLE *) record->record_type_struct;

	if ( old_mathop_variable == (MX_OLD_MATHOP_VARIABLE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_OLD_MATHOP_VARIABLE pointer for record '%s' is NULL.",
			record->name );
	}

	num_records = old_mathop_variable->num_records;

	/* Allocate memory for 'record_array' and 'value_array'. */

	old_mathop_variable->record_array = (MX_RECORD **) malloc(
				num_records * sizeof( MX_RECORD * ) );

	if ( old_mathop_variable->record_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate a %ld element array of "
	"MX_RECORD pointers.", num_records );
	}

	old_mathop_variable->value_array = (double *) malloc(
				num_records * sizeof( double ) );

	if ( old_mathop_variable->value_array == (double *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate a %ld element array of doubles.",
			num_records );
	}

	/* Initialize the contents of 'record_array' and 'value_array'. */

	for ( i = 0; i < num_records; i++ ) {

		name = (old_mathop_variable->record_name_array)[i];

		dependent_record = mx_get_record( record, name );

		(old_mathop_variable->record_array)[i] = dependent_record;

		if ( dependent_record != NULL ) {
			value = 0.0;
		} else {
			/* If the contents of the record_name_array element
			 * does not correspond to a record name, we assume
			 * that it contains a numerical constant instead.
			 *
			 * We can verify this by trying to parse the
			 * string for the numerical constant.
			 */

			value = strtod( name, &endptr );

			if ( *endptr != '\0' ) {
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Elements of the record_name_array for old_mathop record '%s' must be "
	"either MX record names or numerical constants.  The string '%s' "
	"is neither of these things.", record->name, name );
			}
		}
		(old_mathop_variable->value_array)[i] = value;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_old_mathop_dummy_function( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxv_old_mathop_send_variable( MX_VARIABLE *variable )
{
	MX_RECORD *destination_record;
	double destination_value;
	mx_status_type mx_status;

	mx_status = mxv_old_mathop_compute_value( variable,
						&destination_value,
						&destination_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxv_old_mathop_put_value( destination_record,
						destination_value );

	return mx_status;
}

MX_EXPORT mx_status_type
mxv_old_mathop_receive_variable( MX_VARIABLE *variable )
{
	void *value_ptr;
	double *double_ptr;
	double value;
	mx_status_type mx_status;

	mx_status = mxv_old_mathop_compute_value( variable, &value, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_get_variable_pointer( variable->record, &value_ptr );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	double_ptr = (double *) value_ptr;

	*double_ptr = value;

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

static mx_status_type
mxv_old_mathop_get_value( MX_RECORD *record, double *value )
{
	const char fname[] = "mxv_old_mathop_get_value()";

	void *pointer_to_value;
	long num_dimensions, field_type;
	unsigned long ulong_value;
	long long_value;
	int int_value;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}
	if ( value == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The value pointer passed was NULL." );
	}

	*value = 0.0;

	mx_status = MX_SUCCESSFUL_RESULT;

	switch( record->mx_superclass ) {
	case MXR_DEVICE:

		switch( record->mx_class ) {
		case MXC_ANALOG_INPUT:
			mx_status = mx_analog_input_read( record, value );
			break;
		case MXC_ANALOG_OUTPUT:
			mx_status = mx_analog_output_read( record, value );
			break;
		case MXC_DIGITAL_INPUT:
			mx_status = mx_digital_input_read( record,
							&ulong_value );

			*value = (double) ulong_value;
			break;
		case MXC_DIGITAL_OUTPUT:
			mx_status = mx_digital_output_read( record,
							&ulong_value );

			*value = (double) ulong_value;
			break;
		case MXC_MOTOR:
			mx_status = mx_motor_get_position( record, value );
			break;
		case MXC_ENCODER:
			mx_status = mx_encoder_read( record, &long_value );

			*value = (double) long_value;
			break;
		case MXC_SCALER:
			mx_status = mx_scaler_read( record, &long_value );

			*value = (double) long_value;
			break;
		case MXC_TIMER:
			mx_status = mx_timer_read( record, value );
			break;
		case MXC_AMPLIFIER:
			mx_status = mx_amplifier_get_gain( record, value );
			break;
		case MXC_RELAY:
			mx_status = mx_get_relay_status( record, &int_value );
			break;
		default:
			return mx_error( MXE_TYPE_MISMATCH, fname,
		"Device '%s' may not be used by a old_mathop variable record.",
				record->name );
			break;
		}
		break;

	case MXR_VARIABLE:
		/* Find out the size and datatype of the variable. */

		mx_status = mx_get_variable_parameters( record,
					&num_dimensions, NULL,
					&field_type, &pointer_to_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Only 1-dimensional variables are currently supported. */

		if ( num_dimensions != 1 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
			"%ld dimensional variable '%s' may not be used by a "
			"old_mathop variable record.  Only 1-dimensional variables "
			"are supported.", num_dimensions, record->name );
		}

		/* Get the current value of the variable. */

		mx_status = mx_receive_variable( record );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Save the current value of the first element of the
		 * 1-dimensional array.  Any elements after the first
		 * one are ignored.
		 */

		switch( field_type ) {
		case MXFT_CHAR:
			*value = (double) *(char *) pointer_to_value;
			break;
		case MXFT_UCHAR:
			*value = (double) *(unsigned char *) pointer_to_value;
			break;
		case MXFT_SHORT:
			*value = (double) *(short *) pointer_to_value;
			break;
		case MXFT_USHORT:
			*value = (double) *(unsigned short *) pointer_to_value;
			break;
		case MXFT_INT:
			*value = (double) *(int *) pointer_to_value;
			break;
		case MXFT_UINT:
			*value = (double) *(unsigned int *) pointer_to_value;
			break;
		case MXFT_LONG:
			*value = (double) *(long *) pointer_to_value;
			break;
		case MXFT_ULONG:
		case MXFT_HEX:
			*value = (double) *(unsigned long *) pointer_to_value;
			break;
		case MXFT_FLOAT:
			*value = (double) *(float *) pointer_to_value;
			break;
		case MXFT_DOUBLE:
			*value = *(double *) pointer_to_value;
			break;
		default:
			return mx_error( MXE_TYPE_MISMATCH, fname,
			"Variable '%s' is of a type that is not supported "
			"by old_mathop variable records.",
				record->name );
			break;
		}
		break;

	default:
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"Record '%s' may not be used by a old_mathop variable record.  "
		"Only device and variable records may be used.",
			record->name );
		break;
	}

	return mx_status;
}

static mx_status_type
mxv_old_mathop_put_value( MX_RECORD *record, double value )
{
	const char fname[] = "mxv_old_mathop_put_value()";

#if 1
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Writing to a old_mathop variable record is not yet implemented." );
#else
	mx_status_type mx_status;


	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	mx_status = MX_SUCCESSFUL_RESULT;

	switch( record->mx_superclass ) {
	case MXR_DEVICE:

		switch( record->mx_class ) {
		case MXC_ANALOG_INPUT:
		case MXC_ANALOG_OUTPUT:
		case MXC_DIGITAL_INPUT:
		case MXC_DIGITAL_OUTPUT:
		case MXC_MOTOR:
		case MXC_ENCODER:
		case MXC_SCALER:
		case MXC_TIMER:
		case MXC_AMPLIFIER:
		case MXC_RELAY:
		case MXC_MULTICHANNEL_ANALYZER:
		case MXC_MULTICHANNEL_ENCODER:
		case MXC_MULTICHANNEL_SCALER:
			break;
		}
		break;

	case MXR_VARIABLE:

		mx_status = mx_receive_variable( record );
		break;

	default:
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"Record '%s' may not be used by a variable record.  "
		"Only device and variable records may be used.",
			record->name );
		break;
	}

	return mx_status;
#endif
}

static mx_status_type
mxv_old_mathop_compute_value( MX_VARIABLE *variable,
			double *value,
			MX_RECORD **destination_record )
{
	const char fname[] = "mxv_old_mathop_compute_value()";

	MX_RECORD *variable_record;
	MX_OLD_MATHOP_VARIABLE *old_mathop_variable;
	MX_RECORD **record_array;
	double *value_array;
	double local_value, temp_value;
	unsigned long i;
	mx_status_type mx_status;

	if ( variable == (MX_VARIABLE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_VARIABLE pointer passed was NULL." );
	}

	variable_record = variable->record;

	if ( variable_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_VARIABLE structure passed is NULL." );
	}

	old_mathop_variable = (MX_OLD_MATHOP_VARIABLE *)
				variable_record->record_type_struct;

	if ( old_mathop_variable == (MX_OLD_MATHOP_VARIABLE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_OLD_MATHOP_VARIABLE pointer for record '%s' is NULL.",
			variable_record->name );
	}

	if ( old_mathop_variable->num_records <= 0 ) {
		if ( value != NULL ) {
			*value = 0.0;
		}

		return MX_SUCCESSFUL_RESULT;
	}

	record_array = old_mathop_variable->record_array;
	value_array = old_mathop_variable->value_array;

	if ( old_mathop_variable->num_records > 1 ) {

	    if ( (old_mathop_variable->operation_type != MXF_OLD_MATHOP_ADDITION)
	      && (old_mathop_variable->operation_type != MXF_OLD_MATHOP_SUBTRACTION)
	      && (old_mathop_variable->operation_type != MXF_OLD_MATHOP_MULTIPLICATION)
	      && (old_mathop_variable->operation_type != MXF_OLD_MATHOP_DIVISION) )
	    {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"Multiple record math operations are allowed only for "
		"addition, subtraction, multiplication, and division." );
	    } else {

		/* Initialize the value to the first record's value. */

		if ( record_array[0] == (MX_RECORD *) NULL ) {
			local_value = value_array[0];
		} else {
			mx_status = mxv_old_mathop_get_value( record_array[0],
							&local_value );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			value_array[0] = local_value;
		}

		for ( i = 1; i < old_mathop_variable->num_records; i++ ) {

			if ( record_array[i] == (MX_RECORD *) NULL ) {
				local_value = value_array[i];
			} else {
				mx_status = mxv_old_mathop_get_value(
							record_array[i],
							&temp_value );

				if ( mx_status.code != MXE_SUCCESS )
					return mx_status;

				value_array[i] = temp_value;
			}

			switch( old_mathop_variable->operation_type ) {
			case MXF_OLD_MATHOP_ADDITION:
				local_value += temp_value;
				break;
			case MXF_OLD_MATHOP_SUBTRACTION:
				local_value -= temp_value;
				break;
			case MXF_OLD_MATHOP_MULTIPLICATION:
				local_value *= temp_value;
				break;
			case MXF_OLD_MATHOP_DIVISION:
				local_value = mx_divide_safely( local_value,
								temp_value );
				break;
			}
		}
	    }
	} else {
	    if ( record_array[0] == (MX_RECORD *) NULL ) {
		local_value = value_array[0];
	    } else {
		mx_status = mxv_old_mathop_get_value( record_array[0],
							&local_value );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		value_array[0] = local_value;
	    }

	    switch( old_mathop_variable->operation_type ) {
	    case MXF_OLD_MATHOP_ADDITION:
	    case MXF_OLD_MATHOP_SUBTRACTION:
	    case MXF_OLD_MATHOP_MULTIPLICATION:
	    case MXF_OLD_MATHOP_DIVISION:
		break;

	    case MXF_OLD_MATHOP_NEGATION:
		local_value = -local_value;
		break;

	    case MXF_OLD_MATHOP_SQUARE:
		local_value *= local_value;
		break;

	    case MXF_OLD_MATHOP_SQUARE_ROOT:
		if ( local_value < 0.0 ) {
			local_value = 0.0;
		} else {
			local_value = sqrt( local_value );
		}
		break;

	    case MXF_OLD_MATHOP_SINE:
		local_value = sin( local_value );
		break;

	    case MXF_OLD_MATHOP_COSINE:
		local_value = cos( local_value );
		break;

	    case MXF_OLD_MATHOP_TANGENT:
		local_value = tan( local_value );
		break;

	    case MXF_OLD_MATHOP_ARC_SINE:
		if ( local_value > 1.0 ) {
			local_value = 1.0;
		} else if ( local_value < -1.0 ) {
			local_value = -1.0;
		}
		local_value = asin( local_value );
		break;

	    case MXF_OLD_MATHOP_ARC_COSINE:
		if ( local_value > 1.0 ) {
			local_value = 1.0;
		} else if ( local_value < -1.0 ) {
			local_value = -1.0;
		}
		local_value = acos( local_value );
		break;

	    case MXF_OLD_MATHOP_ARC_TANGENT:
		local_value = atan( local_value );
		break;

	    case MXF_OLD_MATHOP_EXPONENTIAL:
		local_value = exp( local_value );
		break;

	    case MXF_OLD_MATHOP_LOGARITHM:
		if ( local_value <= 0.0 ) {
			local_value = - DBL_MAX;
		} else {
			local_value = log( local_value );
		}
		break;

	    case MXF_OLD_MATHOP_LOG10:
		if ( local_value <= 0.0 ) {
			local_value = - DBL_MAX;
		} else {
			local_value = log10( local_value );
		}
		break;

	    default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"Mathop operation type %d is unsupported.",
			old_mathop_variable->operation_type );
		break;
	    }
	}

	if ( value != NULL ) {
		*value = local_value;
	}

	return mx_status;
}

